"""MysqlLegacy module tests."""

import logging
import re
from datetime import datetime
from unittest import mock

import pytest
from ClusterShell.MsgTree import MsgTreeElem
from cumin import Config, nodeset

from spicerack import mysql_legacy
from spicerack.remote import Remote, RemoteExecutionError, RemoteHosts
from spicerack.tests import get_fixture_path
from spicerack.tests.unit.test_remote import mock_cumin

EQIAD_CORE_MASTERS_QUERY = "db10[01-11]"
CODFW_CORE_MASTERS_QUERY = "db20[01-11]"


class TestInstance:
    """Test class for the Instance class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_run_sync = mock.Mock()
        self.config = Config(get_fixture_path("remote", "config.yaml"))
        single = RemoteHosts(self.config, nodeset("single1"), dry_run=False)
        single.run_sync = self.mocked_run_sync
        multi = RemoteHosts(self.config, nodeset("multi1"), dry_run=False)
        multi.run_sync = self.mocked_run_sync
        self.single_instance = mysql_legacy.Instance(single)
        self.multi_instance = mysql_legacy.Instance(multi, name="instance1")

    def test_init_raise(self):
        """It should raise a NotImplementedError exception if more than one host is passed to the constructor."""
        with pytest.raises(NotImplementedError, match="Only single hosts are currently supported"):
            mysql_legacy.Instance(RemoteHosts(self.config, nodeset("host[1-2]")))

    @pytest.mark.parametrize(
        "query, database, kwargs",
        (
            ("SELECT 1 AS test", "", {}),
            ("SELECT 1 AS test FROM dbname", "dbname", {}),
            ("SELECT 1 AS test", "", {"is_safe": True}),
            ("SELECT 1 AS test", "", {"print_progress_bars": True, "print_output": True}),
        ),
    )
    def test_single_run_query_ok(self, query, database, kwargs):
        """It should run the query and return the results."""
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(b"test\n1", parent=MsgTreeElem()))]
        results = self.single_instance.run_query(query, database=database, **kwargs)
        assert [(str(host), msg.message().decode()) for host, msg in results] == [("single1", "test\n1")]
        expected_kwargs = {"print_progress_bars": False, "print_output": False, **kwargs}
        self.mocked_run_sync.assert_called_once_with(
            f'/usr/local/bin/mysql --socket /run/mysqld/mysqld.sock --batch --execute "{query}" {database}'.strip(),
            **expected_kwargs,
        )

    def test_multi_query_raise(self):
        """It should raise a MysqlLegacyError exception if the query execution fails."""
        self.mocked_run_sync.side_effect = RemoteExecutionError(retcode=1, message="error", results=iter(()))
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Failed to run 'invalid' on multi1"):
            self.multi_instance.run_query("invalid")

        self.mocked_run_sync.assert_called_once_with(
            '/usr/local/bin/mysql --socket /run/mysqld/mysqld.instance1.sock --batch --execute "invalid"',
            print_progress_bars=False,
            print_output=False,
        )

    @pytest.mark.parametrize("instance", ("single_instance", "multi_instance"))
    @pytest.mark.parametrize(
        "method, expected",
        (
            ("stop_slave", "STOP SLAVE"),
            ("start_slave", "START SLAVE"),
            ("use_gtid", "CHANGE MASTER TO MASTER_USE_GTID=Slave_pos"),
        ),
    )
    def test_run_method_ok(self, method, expected, instance):
        """It should run the method called and execute the related query."""
        self.mocked_run_sync.return_value = iter(())

        getattr(getattr(self, instance), method)()
        suffix = ".instance1" if instance == "multi_instance" else ""
        self.mocked_run_sync.assert_called_once_with(
            f'/usr/local/bin/mysql --socket /run/mysqld/mysqld{suffix}.sock --batch --execute "{expected}"',
            print_progress_bars=False,
            print_output=False,
        )

    def test_single_show_slave_status_ok(self):
        """It should return the current slave status of the database."""
        status = get_fixture_path("mysql_legacy", "single_show_slave_status.out").read_text().encode()
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(status, parent=MsgTreeElem()))]
        status = self.single_instance.show_slave_status()
        assert status["Master_Host"] == "host1.example.org"
        assert status["Seconds_Behind_Master"] == "0"
        self.mocked_run_sync.assert_called_once_with(
            r'/usr/local/bin/mysql --socket /run/mysqld/mysqld.sock --batch --execute "SHOW SLAVE STATUS\G"',
            is_safe=True,
            print_progress_bars=False,
            print_output=False,
        )

    def test_single_show_slave_status_master(self):
        """It should raise a MysqlLegacyError exception if show slave status is called on a master."""
        self.mocked_run_sync.return_value = iter(())
        with pytest.raises(
            mysql_legacy.MysqlLegacyError, match="SHOW SLAVE STATUS seems to have been executed on a master"
        ):
            self.single_instance.show_slave_status()

    def test_single_show_slave_status_multisource(self):
        """It should raise a MysqlLegacyError exception if show slave status is called on a multisource instance."""
        status = get_fixture_path("mysql_legacy", "single_show_slave_status.out").read_text()
        self.mocked_run_sync.return_value = [
            (nodeset("single1"), MsgTreeElem("".join((status, status)).encode(), parent=MsgTreeElem()))
        ]
        with pytest.raises(NotImplementedError, match="Multisource setup are not implemented"):
            self.single_instance.show_slave_status()

    @pytest.mark.parametrize("instance", ("single_instance", "multi_instance"))
    @pytest.mark.parametrize("method", ("stop", "start", "status", "restart"))
    def test_single_systemctl_action_ok(self, method, instance):
        """It should perform the systemctl action on the mysql process."""
        self.mocked_run_sync.return_value = iter(())
        getattr(getattr(self, instance), f"{method}_mysql")()
        suffix = "@instance1" if instance == "multi_instance" else ""
        kwargs = {}
        if method == "status":
            kwargs["is_safe"] = True
        if method == "start":
            kwargs["print_output"] = True

        self.mocked_run_sync.assert_called_once_with(f"/usr/bin/systemctl {method} mariadb{suffix}.service", **kwargs)

    @pytest.mark.parametrize(
        "instance, path",
        (
            ("single_instance", "/srv/sqldata"),
            ("multi_instance", "/srv/sqldata.instance1"),
        ),
    )
    @pytest.mark.parametrize("skip_confirmation", (False, True))
    @mock.patch("spicerack.mysql_legacy.ask_confirmation")
    def test_clean_data_dir(self, mocked_ask_confirmation, skip_confirmation, instance, path):
        """It should delete the data directory."""
        self.mocked_run_sync.return_value = iter(())
        obj = getattr(self, instance)
        obj.clean_data_dir(skip_confirmation=skip_confirmation)
        self.mocked_run_sync.assert_called_once_with(f"/usr/bin/rm -rf {path}")
        if skip_confirmation:
            mocked_ask_confirmation.assert_not_called()
        else:
            mocked_ask_confirmation.assert_called_once_with(
                f"ATTENTION: destructive action for {obj.host}: /usr/bin/rm -rf {path}. Are you sure to proceed?"
            )

    @pytest.mark.parametrize("instance", ("single_instance", "multi_instance"))
    def test_upgrade(self, instance):
        """It should run the mysql upgrade command."""
        self.mocked_run_sync.return_value = iter(())
        getattr(self, instance).upgrade()
        suffix = ".instance1" if instance == "multi_instance" else ""
        self.mocked_run_sync.assert_called_once_with(
            f"/usr/local/bin/mysql_upgrade --socket /run/mysqld/mysqld{suffix}.sock --force", print_output=True
        )

    def test_get_replication_info_ok(self):
        """It should return a ReplicationInfo instance with the proper data."""
        status = get_fixture_path("mysql_legacy", "single_show_slave_status.out").read_text().encode()
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(status, parent=MsgTreeElem()))]
        info = self.single_instance.get_replication_info()
        self.mocked_run_sync.assert_called_once_with(
            r'/usr/local/bin/mysql --socket /run/mysqld/mysqld.sock --batch --execute "SHOW SLAVE STATUS\G"',
            is_safe=True,
            print_progress_bars=False,
            print_output=False,
        )
        assert info.primary == "host1.example.org"
        assert info.binlog == "host1-bin.001234"
        assert info.position == 123456782

    def test_get_replication_info_raise(self):
        """It should raise a MysqlLegacyError if unable to get the replication information."""
        status = (
            get_fixture_path("mysql_legacy", "single_show_slave_status.out")
            .read_text()
            .replace("Master_Host", "Master_Host_Invalid")
        )
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(status.encode(), parent=MsgTreeElem()))]
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Could not find the replication position"):
            self.single_instance.get_replication_info()

    def test_primary_ok(self):
        """It should return the hostname of the primary host for this host."""
        status = get_fixture_path("mysql_legacy", "single_show_slave_status.out").read_text().encode()
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(status, parent=MsgTreeElem()))]
        assert self.single_instance.primary == "host1.example.org"
        self.mocked_run_sync.assert_called_once()
        # Ensure the caching of the result works
        self.single_instance.primary  # pylint: disable=pointless-statement
        self.mocked_run_sync.assert_called_once()

    @pytest.mark.parametrize("no_content", (False, True))
    def test_primary_raise(self, no_content):
        """It should raise a MysqlLegacyError if there is no primary or is run on a master."""
        if no_content:
            self.mocked_run_sync.return_value = iter(())
        else:
            status = (
                get_fixture_path("mysql_legacy", "single_show_slave_status.out")
                .read_text()
                .replace("Master_Host", "Master_Host_Invalid")
            )
            self.mocked_run_sync.return_value = [
                (nodeset("single1"), MsgTreeElem(status.encode(), parent=MsgTreeElem()))
            ]

        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Unable to retrieve master host"):
            self.single_instance.primary  # pylint: disable=pointless-statement

    def test_prep_src_for_cloning(self):
        """It should run the preparation commands before cloning."""
        status = get_fixture_path("mysql_legacy", "single_show_slave_status.out").read_text().encode()
        node = nodeset("single1")
        self.mocked_run_sync.side_effect = [
            [(node, iter(()))],
            [(node, MsgTreeElem(status, parent=MsgTreeElem()))],
            [(node, iter(()))],
        ]
        info = self.single_instance.prep_src_for_cloning()
        assert info.primary == "host1.example.org"
        calls = self.mocked_run_sync.call_args_list
        assert calls[0][0][0].endswith('STOP SLAVE"')
        assert calls[1][0][0].endswith(r'SHOW SLAVE STATUS\G"')
        assert calls[2][0][0] == "/usr/bin/systemctl stop mariadb.service"

    def test_set_replication_parameters(self):
        """It should set the replication to the given parameters."""
        status = get_fixture_path("mysql_legacy", "single_show_slave_status.out").read_text().encode()
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(status, parent=MsgTreeElem()))]
        info = self.single_instance.get_replication_info()
        self.mocked_run_sync.reset_mock()

        self.multi_instance.set_replication_parameters(replication_info=info, user="user", password="dummy")

        query = self.mocked_run_sync.call_args[0][0]
        for part in (
            "CHANGE MASTER TO",
            "master_host='host1.example.org'",
            "master_port=3306",
            "master_ssl=1",
            "master_log_file='host1-bin.001234'",
            "master_log_pos=123456782",
            "master_user='user'",
            "password='dummy'",
        ):
            assert part in query

    def test_post_clone_reset_with_slave_stopped(self):
        """It should start mysql with slave stopped and reset all slave information."""
        self.mocked_run_sync.side_effect = [[(nodeset("single1"), iter(()))]] * 5
        self.single_instance.post_clone_reset_with_slave_stopped()
        calls = self.mocked_run_sync.call_args_list
        assert calls[0][0][0].endswith("chown -R mysql:mysql /srv/sqldata")
        assert calls[0][0][1].endswith('set-environment MYSQLD_OPTS="--skip-slave-start"')
        assert calls[1][0][0].endswith("systemctl start mariadb.service")
        assert calls[2][0][0].endswith('STOP SLAVE"')
        assert calls[3][0][0].endswith('RESET SLAVE ALL"')

    def test_resume_replication(self):
        """It should start mysql, upgrade it, restart it and resume the replication."""
        self.mocked_run_sync.side_effect = [[(nodeset("single1"), iter(()))]] * 5
        self.single_instance.resume_replication()
        calls = self.mocked_run_sync.call_args_list
        assert calls[0][0][0].endswith('set-environment MYSQLD_OPTS="--skip-slave-start"')
        assert calls[1][0][0].endswith("systemctl start mariadb.service")
        assert calls[2][0][0].endswith("mysql_upgrade --socket /run/mysqld/mysqld.sock --force")
        assert calls[3][0][0].endswith("systemctl restart mariadb.service")
        assert calls[4][0][0].endswith('START SLAVE"')

    @pytest.mark.parametrize("threshold", (0, 0.1234, 0.5))
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_replication_already_ok(self, mocked_sleep, threshold):
        """If the replication is already in sync it should return immediately."""
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(b"lag\n0.1234", parent=MsgTreeElem()))]
        if threshold:
            self.single_instance.wait_for_replication(threshold)
        else:
            threshold = 1.0
            self.multi_instance.wait_for_replication()

        mocked_sleep.assert_not_called()

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_replication_sleep_ok(self, mocked_sleep):
        """If the replication is not yet in sync it should wait until it gets in sync, within the timeout."""
        results = [[(nodeset("single1"), MsgTreeElem(b"lag\n1.1234", parent=MsgTreeElem()))]] * 5
        results.append([(nodeset("single1"), MsgTreeElem(b"lag\n0.1234", parent=MsgTreeElem()))])
        self.mocked_run_sync.side_effect = results
        self.single_instance.wait_for_replication()
        assert mocked_sleep.call_count == 5

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_replication_fail(self, mocked_sleep):
        """If the replication is not in sync within the timeout it should raise a MysqlLegacyReplagError exception."""
        self.mocked_run_sync.side_effect = [
            [(nodeset("single1"), MsgTreeElem(b"lag\n1.1234", parent=MsgTreeElem()))]
        ] * 480
        with pytest.raises(
            mysql_legacy.MysqlLegacyReplagError,
            match=re.escape("Replication lag higher than the threshold (1.0s): 1.1234s"),
        ):
            self.single_instance.wait_for_replication()

        assert mocked_sleep.call_count == 479

    def test_replication_lag_ok(self):
        """It should return the current replication lag as float."""
        self.mocked_run_sync.return_value = [(nodeset("single1"), MsgTreeElem(b"lag\n0.1234", parent=MsgTreeElem()))]
        assert self.single_instance.replication_lag() == pytest.approx(0.1234)

    @pytest.mark.parametrize(
        "results, error_message",
        (
            (None, "Got no output from the replication lag query"),
            (MsgTreeElem(b"invalid", parent=MsgTreeElem()), "Unable to parse replication lag from: ['invalid']"),
            (
                MsgTreeElem(b"invalid\nnot-a-float", parent=MsgTreeElem()),
                "Unable to parse replication lag from: ['invalid', 'not-a-float']",
            ),
        ),
    )
    def test_replication_lag_fail(self, results, error_message):
        """It should raise a MysqlLegacyError if there is no output or the output cannot be parsed as lag."""
        self.mocked_run_sync.return_value = [(nodeset("single1"), results)] if results is not None else []
        with pytest.raises(mysql_legacy.MysqlLegacyError, match=re.escape(error_message)):
            self.single_instance.replication_lag()


class TestMysqlLegacyRemoteHosts:
    """Test class for the MysqlLegacyRemoteHosts class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_run_sync = mock.Mock()
        self.config = Config(get_fixture_path("remote", "config.yaml"))

        remote_hosts = RemoteHosts(self.config, nodeset("host[1-9]"), dry_run=False)
        remote_hosts.run_sync = self.mocked_run_sync
        self.mysql_remote_hosts = mysql_legacy.MysqlLegacyRemoteHosts(remote_hosts)

        remote_host = RemoteHosts(self.config, nodeset("host1"), dry_run=False)
        remote_host.run_sync = self.mocked_run_sync
        self.mysql_remote_host = mysql_legacy.MysqlLegacyRemoteHosts(remote_host)

    def test_run_query(self):
        """Calling run_query() should run the given query in the target hosts."""
        self.mocked_run_sync.return_value = [(nodeset("host[1-9]"), MsgTreeElem(b"result", parent=MsgTreeElem()))]
        results = self.mysql_remote_hosts.run_query("SELECT 'result'")
        assert [(str(host), msg.message().decode()) for host, msg in results] == [("host[1-9]", "result")]
        self.mocked_run_sync.assert_called_once_with(
            "/usr/local/bin/mysql --skip-ssl --skip-column-names --batch -e \"SELECT 'result'\"",
            print_progress_bars=False,
            print_output=False,
        )

    def test_list_host_instances_no_instance(self):
        """It should return an empty list if there are no instances on the host."""
        self.mocked_run_sync.return_value = iter(())
        instances = self.mysql_remote_host.list_hosts_instances()
        assert instances == []  # pylint: disable=use-implicit-booleaness-not-comparison

    def test_list_host_instances_single(self):
        """It should return a list with just one Instance object for the single instance."""
        service = b"mariadb.service loaded active running mariadb database server"
        self.mocked_run_sync.return_value = [(nodeset("host1"), MsgTreeElem(service, parent=MsgTreeElem()))]
        instances = self.mysql_remote_host.list_hosts_instances()
        assert len(instances) == 1
        assert isinstance(instances[0], mysql_legacy.Instance)
        assert str(instances[0].host) == "host1"
        assert instances[0].name == ""
        self.mocked_run_sync.assert_called_once_with(
            "/usr/bin/systemctl --no-pager --type=service --plain --no-legend  list-units 'mariadb*'",
            is_safe=True,
            print_progress_bars=False,
            print_output=False,
        )

    def test_list_host_instances_multi(self):
        """It should return a list with all the instances of a multi-instance host."""
        services = b"\n".join(
            [
                b"mariadb@s1.service loaded active running mariadb database server",
                b"mariadb@s2.service loaded active running mariadb database server",
                b"mariadb-spurious.service loaded active running mariadb database server",
            ]
        )
        self.mocked_run_sync.return_value = [(nodeset("host1"), MsgTreeElem(services, parent=MsgTreeElem()))]
        instances = self.mysql_remote_host.list_hosts_instances()
        assert len(instances) == 2
        for instance in instances:
            assert isinstance(instance, mysql_legacy.Instance)
            assert str(instance.host) == "host1"

        assert instances[0].name == "s1"
        assert instances[1].name == "s2"

    def test_list_host_instances_not_single_host(self):
        """It should raise a NotImplementedError if the MysqlLegacyRemoteHosts instance has multiple hosts."""
        with pytest.raises(NotImplementedError, match="Only single host are supported at this time"):
            self.mysql_remote_hosts.list_hosts_instances()

    def test_list_host_instances_grouped(self):
        """It should raise a NotImplementedError if the grouped parameter is set to True."""
        with pytest.raises(NotImplementedError, match="Grouping and parallelization are not supported at this time"):
            self.mysql_remote_host.list_hosts_instances(group=True)


class TestMysqlLegacy:
    """MysqlLegacy class tests."""

    @mock.patch("spicerack.remote.transports", autospec=True)
    def setup_method(self, _, mocked_transports):
        """Initialize the test environment for MysqlLegacy."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path("remote", "config.yaml"))
        self.mocked_transports = mocked_transports
        self.mocked_remote = mock.MagicMock(spec_set=Remote)
        self.mysql = mysql_legacy.MysqlLegacy(self.mocked_remote, dry_run=False)

    def test_get_dbs(self):
        """It should return and instance of MysqlLegacyRemoteHosts for the matching hosts."""
        self.mysql.get_dbs("query")
        self.mocked_remote.query.assert_called_once_with("query")

    @pytest.mark.parametrize(
        "kwargs, query, match",
        (
            ({}, "A:db-core", "db10[01-99],db20[01-99]"),
            ({"datacenter": "eqiad"}, "A:db-core and A:eqiad", "db10[01-99]"),
            (
                {"section": "s1"},
                "A:db-core and A:db-section-s1",
                "db10[01-10],db20[01-10]",
            ),
            (
                {"replication_role": "master"},
                "A:db-core and A:db-role-master",
                ",".join([EQIAD_CORE_MASTERS_QUERY, CODFW_CORE_MASTERS_QUERY]),
            ),
            (
                {"datacenter": "eqiad", "section": "s1"},
                "A:db-core and A:eqiad and A:db-section-s1",
                "db10[01-10]",
            ),
            (
                {"datacenter": "eqiad", "replication_role": "master"},
                "A:db-core and A:eqiad and A:db-role-master",
                EQIAD_CORE_MASTERS_QUERY,
            ),
            (
                {"datacenter": "eqiad", "replication_role": "master", "excludes": ("s2",)},
                "A:db-core and A:eqiad and not A:db-section-s2 and A:db-role-master",
                "db10[01-10]",
            ),
            (
                {"section": "s1", "replication_role": "master"},
                "A:db-core and A:db-section-s1 and A:db-role-master",
                "db1001,db2001",
            ),
            (
                {"datacenter": "eqiad", "section": "s1", "replication_role": "master"},
                "A:db-core and A:eqiad and A:db-section-s1 and A:db-role-master",
                "db1001",
            ),
        ),
    )
    def test_get_core_dbs_ok(self, kwargs, query, match):
        """It should return the right DBs based on the parameters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset(match))
        self.mysql.get_core_dbs(**kwargs)
        self.mocked_remote.query.assert_called_once_with(query)

    @pytest.mark.parametrize(
        "kwargs",
        (
            {"datacenter": "invalid"},
            {"section": "invalid"},
            {"replication_role": "invalid"},
            {"excludes": ("invalid",)},
        ),
    )
    def test_get_core_dbs_fail(self, kwargs):
        """It should raise MysqlLegacyError if called with invalid parameters."""
        key = list(kwargs.keys())[0]
        message = f"Got invalid {key}"
        with pytest.raises(mysql_legacy.MysqlLegacyError, match=message):
            self.mysql.get_core_dbs(**kwargs)

        assert not self.mocked_remote.query.called

    def test_get_core_dbs_fail_sanity_check(self):
        """It should raise MysqlLegacyError if matching an invalid number of hosts when looking for masters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Matched 1 masters, expected 11"):
            self.mysql.get_core_dbs(datacenter="eqiad", replication_role="master")

        assert self.mocked_remote.query.called

    @pytest.mark.parametrize("mode, value", (("readonly", b"1"), ("readwrite", b"0")))
    def test_set_core_masters_readonly(self, mode, value, caplog):
        """It should set the masters as read-only/read-write."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(self.mocked_transports, 0, retvals=[[("db10[01-11]", value)]])
        with caplog.at_level(logging.DEBUG):
            getattr(self.mysql, "set_core_masters_" + mode)("eqiad")
        assert "SET GLOBAL read_only=" + value.decode() in caplog.text

    @pytest.mark.parametrize("readonly, reply", ((True, b"1"), (False, b"0")))
    def test_verify_core_masters_readonly_ok(self, readonly, reply, caplog):
        """Should verify that the masters have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(self.mocked_transports, 0, retvals=[[("db10[01-11]", reply)]])
        with caplog.at_level(logging.DEBUG):
            self.mysql.verify_core_masters_readonly("eqiad", readonly)
        assert "SELECT @@global.read_only" in caplog.text

    def test_verify_core_masters_readonly_fail(self):
        """Should raise MysqlLegacyError if some masters do not have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db10[01-11]"))
        mock_cumin(
            self.mocked_transports,
            0,
            retvals=[[("db1001", b"0"), ("db10[02-11]", b"1")]],
        )
        with pytest.raises(
            mysql_legacy.MysqlLegacyError,
            match="Verification failed that core DB masters",
        ):
            self.mysql.verify_core_masters_readonly("eqiad", True)

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_ok(self, mocked_sleep):
        """Should check that all core masters are in sync with the master in the other DC."""
        hosts = nodeset(EQIAD_CORE_MASTERS_QUERY)
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, nodeset(host)) for host in hosts] * 2
        retvals = [[(host, b"2018-09-06T10:00:00.000000")] for host in hosts]  # first heartbeat
        retvals += [[(host, b"2018-09-06T10:00:01.000000")] for host in hosts]  # second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        self.mysql.check_core_masters_in_sync("eqiad", "codfw")
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_fail_heartbeat(self, mocked_sleep):
        """Should raise MysqlLegacyError if unable to get the heartbeat from the current master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Unable to get heartbeat from master"):
            self.mysql.check_core_masters_in_sync("eqiad", "codfw")
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_in_sync_not_in_sync(self, mocked_sleep):
        """Should raise MysqlLegacyError if a master is not in sync with the one in the other DC."""
        hosts = nodeset(EQIAD_CORE_MASTERS_QUERY)
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, nodeset(host)) for host in hosts] + [
            RemoteHosts(self.config, nodeset("db1001"))
        ] * 3
        retvals = [[(host, b"2018-09-06T10:00:00.000000")] for host in hosts]  # first heartbeat
        retvals += [[("db1001", b"2018-09-06T10:00:00.000000")]] * 3  # 3 failed retries of second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        with pytest.raises(
            mysql_legacy.MysqlLegacyError,
            match=r"Heartbeat from master db1001 for section .* not yet in sync",
        ):
            self.mysql.check_core_masters_in_sync("eqiad", "codfw")

        assert mocked_sleep.called

    def test_get_core_masters_heartbeats_wrong_data(self):
        """Should raise MysqlLegacyError if unable to convert the heartbeat into a datetime."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(
            self.mocked_transports,
            0,
            retvals=[[("db1001", b"2018-09-06-10:00:00.000000")]],
        )
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Unable to convert heartbeat"):
            self.mysql.get_core_masters_heartbeats("eqiad", "codfw")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_check_core_masters_heartbeats_fail(self, mocked_sleep):
        """Should raise MysqlLegacyError if unable to get the heartbeat from the master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, nodeset("db1001"))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql_legacy.MysqlLegacyError, match="Unable to get heartbeat from master"):
            self.mysql.check_core_masters_heartbeats("eqiad", "codfw", {"s1": datetime.utcnow()})

        assert mocked_sleep.called



"""Service Module Tests."""
from ipaddress import ip_address
from unittest import mock

import pytest
from wmflib.config import load_yaml_config

from spicerack import Alertmanager, service
from spicerack.administrative import Reason
from spicerack.dnsdisc import DiscoveryError
from spicerack.tests import get_fixture_path


def test_service_discovery_no_records():
    """It should raise a DiscoveryRecordNotFoundError exception if no records are present."""
    discovery = service.ServiceDiscovery([])
    with pytest.raises(service.DiscoveryRecordNotFoundError, match="No DNS Discovery record present."):
        discovery.get()


class TestCatalog:
    """Test class for the Catalog class."""

    def setup_method(self):
        """Initialize the tests."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_confctl = mock.MagicMock()
        self.authdns_servers = load_yaml_config(get_fixture_path("discovery", "authdns.yaml"))

        catalog = load_yaml_config(get_fixture_path("service", "service.yaml"))
        alertmanager = Alertmanager(alertmanager_urls=("https://alertmanager-eqiad.wikimedia.example",))
        self.catalog = service.Catalog(
            catalog, alertmanager=alertmanager, confctl=self.mocked_confctl, authdns_servers=self.authdns_servers
        )

    def test_init(self):
        """It should instantiate a Catalog instance properly."""
        assert isinstance(self.catalog, service.Catalog)

    def test_service_names(self):
        """It should return the list of all service names."""
        assert self.catalog.service_names == ["service1", "service_no_lvs", "service2", "service3", "service4"]

    def test_get_instance(self):
        """It should return a Service instance for the given service."""
        assert isinstance(self.catalog.get("service1"), service.Service)

    def test_get_raise(self):
        """It should raise a ServiceNotFoundError if the service is not found."""
        with pytest.raises(service.ServiceNotFoundError, match="Service nonexistent was not found"):
            self.catalog.get("nonexistent")

    def test_iter(self):
        """It should allow to iterate over the catalog and get a Service instance for each service."""
        assert [service.name for service in self.catalog] == self.catalog.service_names

    def test_len(self):
        """It should return the number of services in the catalog."""
        assert len(self.catalog) == 5


class TestService:
    """Test class for the Service class."""

    def setup_method(self):
        """Initialize the tests."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_confctl = mock.MagicMock()
        self.authdns_servers = load_yaml_config(get_fixture_path("discovery", "authdns.yaml"))

        catalog = load_yaml_config(get_fixture_path("service", "service.yaml"))
        alertmanager = Alertmanager(alertmanager_urls=("https://alertmanager-eqiad.wikimedia.example",), dry_run=False)
        self.catalog = service.Catalog(
            catalog,
            alertmanager=alertmanager,
            confctl=self.mocked_confctl,
            authdns_servers=self.authdns_servers,
            dry_run=False,
        )

        self.service1 = self.catalog.get("service1")
        self.service2 = self.catalog.get("service2")
        self.service3 = self.catalog.get("service3")
        self.service_no_lvs = self.catalog.get("service_no_lvs")
        self.subnets = {"eqiad": ip_address("10.10.0.1"), "codfw": ip_address("10.20.0.1")}

    @pytest.mark.parametrize(
        "attrs, obj_type",
        (
            ((), service.Service),
            (("discovery",), service.ServiceDiscovery),
            (("discovery", "get"), service.ServiceDiscoveryRecord),
            (("ip",), service.ServiceIPs),
            (("lvs",), service.ServiceLVS),
            (("lvs", "conftool"), service.ServiceLVSConftool),
            (("monitoring",), service.ServiceMonitoring),
            (("monitoring", "sites"), service.ServiceMonitoringHostnames),
        ),
    )
    def test_instances_service1(self, attrs, obj_type):
        """The service properties should be an instance of the given class."""
        obj = self.service1
        for attr in attrs:
            obj = getattr(obj, attr)
            if callable(obj):
                obj = obj()

        assert isinstance(obj, obj_type)

    @pytest.mark.parametrize(
        "attr, obj_type",
        (
            ("", service.Service),
            ("ip", service.ServiceIPs),
        ),
    )
    def test_instances_service2(self, attr, obj_type):
        """The service properties should be an instance of the given class."""
        obj = self.service2
        if attr:
            obj = getattr(obj, attr)

        assert isinstance(obj, obj_type)

    @pytest.mark.parametrize(
        "attrs, expected",
        (
            (("name",), "service1"),
            (("description",), "Service1 description"),
            (("discovery", "get", "dnsdisc"), "service1"),
            (("discovery", "get", "active_active"), True),
            (("encryption",), True),
            (("ip", "all"), [ip_address("10.2.1.1"), ip_address("10.2.2.1")]),
            (("ip", "sites"), ["codfw", "eqiad"]),
            (("lvs", "depool_threshold"), ".5"),
            (("lvs", "enabled"), True),
            (("lvs", "lvs_class"), "low-traffic"),
            (("lvs", "bgp"), True),
            (("lvs", "protocol"), "tcp"),
            (("lvs", "scheduler"), "wrr"),
            (("lvs", "conftool", "cluster"), "cluster1"),
            (("lvs", "conftool", "service"), "service1"),
            (("monitoring", "check_command"), "check_https_lvs_on_port!service1.discovery.wmnet!443!/health"),
            (("monitoring", "contact_group"), ""),
            (("monitoring", "notes_url"), ""),
            (("monitoring", "sites", "all"), ["service1.svc.codfw.wmnet", "service1.svc.eqiad.wmnet"]),
            (("monitoring", "sites", "sites"), ["codfw", "eqiad"]),
            (("page",), False),
            (("port",), 443),
            (("probes",), [{"type": "http", "path": "/health"}]),
            (("sites",), ["codfw", "eqiad"]),
            (("state",), "production"),
        ),
    )
    def test_property_service1(self, attrs, expected):
        """It should access the Service 1 properties and return their value."""
        obj = self.service1
        for attr in attrs:
            obj = getattr(obj, attr)
            if callable(obj):
                obj = obj()

        assert obj == expected

    @pytest.mark.parametrize(
        "attr, expected",
        (
            ("discovery", None),
            ("lvs", None),
            ("monitoring", None),
            ("page", True),
        ),
    )
    def test_property_service2(self, attr, expected):
        """It should access the Service 2 properties missing and return their default values."""
        assert getattr(self.service2, attr) == expected

    @pytest.mark.parametrize(
        "site, label, expected",
        (
            ("codfw", "", ip_address("10.2.1.1")),
            ("codfw", "default", ip_address("10.2.1.1")),
            ("eqiad", "", ip_address("10.2.2.1")),
            ("nonexistent", "", None),
            ("nonexistent", "nonexistent", None),
        ),
    )
    def test_ip_sites_get(self, site, label, expected):
        """It should return the hostname/FQDN of the given site."""
        if label:
            assert self.service1.ip.get(site, label=label) == expected
        else:
            assert self.service1.ip.get(site) == expected

    @pytest.mark.parametrize(
        "site, expected",
        (
            ("codfw", "service1.svc.codfw.wmnet"),
            ("eqiad", "service1.svc.eqiad.wmnet"),
            ("nonexistent", ""),
        ),
    )
    def test_monitoring_sites_get(self, site, expected):
        """It should return the hostname/FQDN of the given site."""
        assert self.service1.monitoring.sites.get(site) == expected

    def test_discovery_multiple_records(self):
        """It should allow to access all of them."""
        dnsdisc = ["service3_a", "service3_b"]
        active = [True, False]
        discovery = self.service3.discovery
        assert len(discovery) == len(dnsdisc)
        for i, record in enumerate(discovery):
            assert record.dnsdisc == dnsdisc[i]
            assert record.active_active == active[i]
            assert self.service3.discovery.get(record.dnsdisc) is record

    def test_discovery_get_multiple_records_no_name(self):
        """It should raise a TooManyDiscoveryRecordsError exception if name is not passed with multiple records."""
        with pytest.raises(
            service.TooManyDiscoveryRecordsError, match="There are 2 DNS Discovery records but dnsdisc was not set."
        ):
            self.service3.discovery.get()

    def test_discovery_get_multiple_records_not_found(self):
        """It should raise a DiscoveryRecordNotFoundError exception if no matching record is found."""
        with pytest.raises(
            service.DiscoveryRecordNotFoundError, match="Unable to find DNS Discovery record nonexistent."
        ):
            self.service3.discovery.get("nonexistent")

    def test_discovery_get_multiple_records_too_many(self):
        """It should raise a TooManyDiscoveryRecordsError exception if more than one match is found."""
        with pytest.raises(
            service.TooManyDiscoveryRecordsError,
            match="There are 2 DNS Discovery records matching name service_no_lvs.",
        ):
            self.service_no_lvs.discovery.get("service_no_lvs")

    def test_discovery_depool(self):
        """It should depool the correct service from DNS Discovery from the given site."""
        self.service1.discovery.depool("codfw")
        self.mocked_confctl.set_and_verify.assert_called_once_with("pooled", False, dnsdisc="(service1)", name="codfw")

    def test_discovery_pool(self):
        """It should pool the correct service from DNS Discovery from the given site."""
        self.service1.discovery.pool("codfw")
        self.mocked_confctl.set_and_verify.assert_called_once_with("pooled", True, dnsdisc="(service1)", name="codfw")

    def test_discovery_pooled_ok(self):
        """It should depool, give control and repool the correct service from DNS Discovery from the given site."""
        with self.service1.discovery.depooled("codfw"):
            self.mocked_confctl.set_and_verify.assert_called_once_with(
                "pooled", False, dnsdisc="(service1)", name="codfw"
            )
            self.mocked_confctl.reset_mock()
        self.mocked_confctl.set_and_verify.assert_called_once_with("pooled", True, dnsdisc="(service1)", name="codfw")

    def test_discovery_pooled_raise(self):
        """It should not repool the given service if an exception is raised."""
        with pytest.raises(RuntimeError, match="some error"):
            with self.service1.discovery.depooled("codfw"):
                self.mocked_confctl.set_and_verify.assert_called_once_with(
                    "pooled", False, dnsdisc="(service1)", name="codfw"
                )
                self.mocked_confctl.reset_mock()
                raise RuntimeError("some error")
        self.mocked_confctl.set_and_verify.assert_not_called()

    def test_discovery_pooled_raise_repool(self):
        """It should repool the given service if an exception is raised but repool_on_error is True."""
        with pytest.raises(RuntimeError, match="some error"):
            with self.service1.discovery.depooled("codfw", repool_on_error=True):
                self.mocked_confctl.set_and_verify.assert_called_once_with(
                    "pooled", False, dnsdisc="(service1)", name="codfw"
                )
                self.mocked_confctl.reset_mock()
                raise RuntimeError("some error")
        self.mocked_confctl.set_and_verify.assert_called_once_with("pooled", True, dnsdisc="(service1)", name="codfw")

    def test_downtime(self, requests_mock):
        """It should downtime the service on Alertmanager for the given site."""
        requests_mock.post("/api/v2/silences", json={"silenceID": "a123"})
        reason = Reason("Maintenance", "user", "host")
        downtime_id = self.service1.downtime("codfw", reason)
        assert downtime_id == "a123"
        request_json = requests_mock.last_request.json()
        assert request_json["matchers"] == [
            {"name": "site", "value": "codfw", "isRegex": False},
            {"name": "job", "value": r"^probes/.*", "isRegex": True},
            {"name": "instance", "value": "^(service1:443)$", "isRegex": True},
        ]

    def test_downtime_raise(self):
        """It should raise a ServiceError when trying to downtime a service in a site where is not present."""
        reason = Reason("Maintenance", "user", "host")
        with pytest.raises(
            service.ServiceError,
            match=r"Service service1 is not present in site nonexistent. Available sites are \['codfw', 'eqiad'\]",
        ):
            self.service1.downtime("nonexistent", reason)

    def test_downtimed(self, requests_mock):
        """It should allow to perform actions while the service is downtimed in the given site."""
        requests_mock.post("/api/v2/silences", json={"silenceID": "a123"})
        requests_mock.delete("/api/v2/silence/a123")
        reason = Reason("Maintenance", "user", "host")
        with self.service1.downtimed("codfw", reason):
            request_json = requests_mock.last_request.json()
            assert request_json["matchers"] == [
                {"name": "site", "value": "codfw", "isRegex": False},
                {"name": "job", "value": r"^probes/.*", "isRegex": True},
                {"name": "instance", "value": "^(service1:443)$", "isRegex": True},
            ]
        assert requests_mock.call_count == 2
        assert requests_mock.request_history[-1].method == "DELETE"

    def test_is_pooled_in(self):
        """A discovery record correctly reports if a datacenter is pooled or not."""
        with mock.patch(
            "spicerack.service.Discovery.active_datacenters", new_callable=mock.PropertyMock
        ) as act_dc_mock:
            act_dc_mock.return_value = {"service1": ["eqiad"]}
            record = self.service1.discovery.records[0]
            assert record.is_pooled_in("eqiad")
            assert record.is_pooled_in("codfw") is False

    @pytest.mark.parametrize(
        "active_dcs, is_ok",
        [(["codfw"], True), (["codfw", "eqiad"], False), (["eqiad"], False)],
    )
    def test_check_dns_state(self, active_dcs, is_ok):
        """It should raise a DiscoveryStateError if the dns responses do not correspond to the etcd discovery state."""
        record = self.service1.discovery.records[0]
        # Always return the IP of codfw
        record.instance.resolve_with_client_ip = mock.MagicMock(
            return_value={"eqiad": ip_address("10.2.1.1"), "codfw": ip_address("10.2.1.1")}
        )
        # Also set the pooled state
        with mock.patch(
            "spicerack.service.Discovery.active_datacenters", new_callable=mock.PropertyMock
        ) as act_dc_mock:
            act_dc_mock.return_value = {"service1": active_dcs}
            if is_ok:
                self.service1.check_dns_state(self.subnets)
            else:
                with pytest.raises(service.DiscoveryStateError):
                    self.service1.check_dns_state(self.subnets, tries=1)

    def test_check_dns_state_resolve_error(self):
        """It should raise a DiscoveryStateError if the name resolution fails."""
        record = self.service1.discovery.records[0]
        record.instance.resolve_with_client_ip = mock.MagicMock(side_effect=DiscoveryError("pinkunicorn"))
        with pytest.raises(service.DiscoveryStateError, match="pinkunicorn"):
            self.service1.check_dns_state(self.subnets, tries=1)

    def test_check_dns_state_no_discovery(self):
        """It should return if there is no discovery record."""
        assert self.service2.check_dns_state(self.subnets) is None

    def test_check_dns_state_wrong_record(self):
        """It should raise a DiscoveryNotFoundError if a bad record is passed."""
        with pytest.raises(
            service.DiscoveryRecordNotFoundError, match="Unable to find DNS Discovery record nonexistent."
        ):
            self.service1.check_dns_state(self.subnets, "nonexistent", tries=1)

    def test_check_dns_state_wrong_tries(self):
        """It should raise a ValueError if a wrong tries value is passed."""
        with pytest.raises(ValueError, match="The tries argument must be a positive integer"):
            self.service1.check_dns_state(self.subnets, tries=0)

    def test_check_dns_state_too_many_records(self):
        """It should raise a TooManyDiscoveryRecordsError if more than one record can be selected."""
        with pytest.raises(
            service.TooManyDiscoveryRecordsError,
            match="There are 2 DNS Discovery records matching name service_no_lvs.",
        ):
            self.service_no_lvs.check_dns_state(self.subnets, "service_no_lvs", tries=1)



"""Netbox module tests."""

import ipaddress
import logging
from copy import deepcopy
from datetime import datetime
from io import BytesIO
from pathlib import Path
from unittest import mock

import pytest
import requests
from packaging import version

from spicerack import redfish

ACCOUNTS_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#ManagerAccountCollection.ManagerAccountCollection",
    "@odata.id": "/redfish/v1/AccountService/Accounts",
    "@odata.type": "#ManagerAccountCollection.ManagerAccountCollection",
    "Description": "BMC User Accounts Collection",
    "Members": [
        {"@odata.id": "/redfish/v1/AccountService/Accounts/1"},
        {"@odata.id": "/redfish/v1/AccountService/Accounts/2"},
        {"@odata.id": "/redfish/v1/AccountService/Accounts/3"},
    ],
    "Members@odata.count": 3,
    "Name": "Accounts Collection",
}
ACCOUNT_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#ManagerAccount.ManagerAccount",
    "@odata.id": "/redfish/v1/AccountService/Accounts/{user_id}",
    "@odata.type": "#ManagerAccount.v1_5_0.ManagerAccount",
    "AccountTypes": ["Redfish", "SNMP", "OEM"],
    "Description": "User Account",
    "Enabled": True,
    "Id": "{user_id}",
    "Links": {"Role": {"@odata.id": "/redfish/v1/AccountService/Roles/Administrator"}},
    "Locked": False,
    "Name": "User Account",
    "OEMAccountTypes": ["IPMI", "SOL", "WSMAN", "UI", "RACADM"],
    "Password": None,
    "PasswordChangeRequired": False,
    "RoleId": "Administrator",
    "SNMP": {
        "AuthenticationKey": None,
        "AuthenticationKeySet": True,
        "AuthenticationProtocol": "HMAC_SHA96",
        "EncryptionKey": None,
        "EncryptionKeySet": False,
        "EncryptionProtocol": "CFB128_AES128",
    },
    "UserName": "username",
}
DELL_SCP = {
    "SystemConfiguration": {
        "Comments": [{"Comment": "First comment"}],
        "Components": [
            {
                "Attributes": [
                    {
                        "Comment": "Read and Write",
                        "Name": "Some.Attribute.2",
                        "Set On Import": "False",
                        "Value": "value",
                    },
                    {
                        "Comment": "Read and Write",
                        "Name": "Some.Attribute.1",
                        "Set On Import": "False",
                        "Value": "value",
                    },
                ],
                "FQDD": "Some.Component.2",
            },
            {
                "Attributes": [
                    {
                        "Comment": "Read and Write",
                        "Name": "Some.Attribute.1",
                        "Set On Import": "False",
                        "Value": "value",
                    },
                ],
                "FQDD": "Some.Component.1",
            },
            {
                "Attributes": [
                    {
                        "Comment": "Read and Write",
                        "Name": "Comma.Separated.List.1",
                        "Set On Import": "False",
                        "Value": "value1,value2",
                    },
                    {
                        "Comment": "Read and Write",
                        "Name": "Comma.Space.Separated.List.1",
                        "Set On Import": "False",
                        "Value": "value1, value2",
                    },
                ],
                "FQDD": "List.Component.1",
            },
        ],
        "Model": "PowerEdge R440",
        "ServiceTag": "12ABC34",
        "TimeStamp": "Thu Dec  9 09:32:06 2021",
    },
}
DELL_TASK_REPONSE = {
    "@odata.context": "/redfish/v1/$metadata#Task.Task",
    "@odata.id": "/redfish/v1/TaskService/Tasks/JID_1234567890",
    "@odata.type": "#Task.v1_1_1.Task",
    "Description": "Server Configuration and other Tasks running on iDRAC are listed here",
    "EndTime": "TIME_NA",
    "Id": "JID_1234567890",
    "Messages": [
        {
            "Oem": {
                "Custom": "Structure",
            },
        },
        {
            "Message": "Exporting Server Configuration Profile.",
            "MessageArgs": [],
            "MessageArgs@odata.count": 0,
            "MessageId": "SYS057",
        },
        {
            "Message": "Successfully exported Server Configuration Profile",
            "MessageArgs": [],
            "MessageArgs@odata.count": 0,
            "MessageID": "SYS043",
        },
    ],
    "Messages@odata.count": 1,
    "Name": "Export: Server Configuration Profile",
    "StartTime": "2021-12-09T14:36:29-06:00",
    "TaskState": "Running",
    "TaskStatus": "OK",
}
DELL_TASK_REPONSE_EPOC = deepcopy(DELL_TASK_REPONSE)
DELL_TASK_REPONSE_EPOC["EndTime"] = "1969-12-31T18:00:00-06:00"
DELL_TASK_REPONSE_BAD_TIME = deepcopy(DELL_TASK_REPONSE)
DELL_TASK_REPONSE_BAD_TIME["EndTime"] = "bad value"
MANAGER_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#Manager.Manager",
    "@odata.id": "/redfish/v1/Managers/Testing_oob.1",
    "@odata.type": "#Manager.v1_12_0.Manager",
    "Actions": {
        "#Manager.Reset": {
            "ResetType@Redfish.AllowableValues": ["GracefulRestart"],
            "target": "/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Manager.Reset",
        },
        "#Manager.ResetToDefaults": {
            "ResetType@Redfish.AllowableValues": ["ResetAll", "PreserveNetworkAndUsers"],
            "target": "/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Manager.ResetToDefaults",
        },
    },
    "DateTime": "2023-01-30T14:23:51-06:00",
    "DateTimeLocalOffset": "-06:00",
    "Description": "BMC",
    "EthernetInterfaces": {"@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/EthernetInterfaces"},
    "FirmwareVersion": "6.00.30.00",
    "HostInterfaces": {"@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/HostInterfaces"},
    "Id": "iDRAC.Embedded.1",
    "LastResetTime": "2022-11-17T17:56:11-06:00",
    "LogServices": {"@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices"},
    "ManagerType": "BMC",
    "Model": "14G Monolithic",
    "Name": "Manager",
    "NetworkProtocol": {"@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/NetworkProtocol"},
    "PowerState": "On",
    "SerialConsole": {
        "ConnectTypesSupported": [],
        "ConnectTypesSupported@odata.count": 0,
        "MaxConcurrentSessions": 0,
        "ServiceEnabled": False,
    },
    "SerialInterfaces": {"@odata.id": "/redfish/v1/Managers/iDRAC.Embedded.1/SerialInterfaces"},
    "Status": {"Health": "OK", "State": "Enabled"},
}
MANAGER_RESPONSE_V3 = deepcopy(MANAGER_RESPONSE)
MANAGER_RESPONSE_V3["FirmwareVersion"] = "3.0.0.0"
MANAGER_RESPONSE_BAD = deepcopy(MANAGER_RESPONSE)
MANAGER_RESPONSE_BAD["Model"] = "Foobar"
LCLOG_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#LogEntryCollection.LogEntryCollection",
    "@odata.id": "/redfish/v1/Managers/Testing_oob.1/LogServices/Lclog/Entries",
    "@odata.type": "#LogEntryCollection.LogEntryCollection",
    "Description": "LC Logs for this manager",
    "Members": [
        {
            "@odata.id": "/redfish/v1/Managers/Testing_oob.1/LogServices/Lclog/Entries/1735",
            "@odata.type": "#LogEntry.v1_6_1.LogEntry",
            "Created": "2022-06-22T17:01:17-05:00",
            "Description": "Log Entry 1735",
            "EntryType": "Oem",
            "Id": "1735",
            "Links": {"OriginOfCondition": {"@odata.id": "/redfish/v1/Managers/Testing_oob.1"}},
            "Message": "Successfully logged in using root, from " "10.192.32.49 and REDFISH.",
            "MessageArgs": ["root", "10.192.32.49", "REDFISH"],
            "MessageArgs@odata.count": 3,
            "MessageId": "USR0030",
            "Name": "Log Entry 1735",
            "Oem": {
                "Dell": {
                    "@odata.type": "#DellLCLogEntry.v1_0_0.DellLCLogEntry",
                    "Category": "Audit",
                    "Comment": None,
                    "LastUpdatedByUser": None,
                }
            },
            "OemRecordFormat": "Dell",
            "Severity": "OK",
        },
        {
            "@odata.id": "/redfish/v1/Managers/Testing_oob.1/LogServices/Lclog/Entries/1734",
            "@odata.type": "#LogEntry.v1_6_1.LogEntry",
            "Created": "2022-06-22T16:58:17-05:00",
            "Description": "Log Entry 1734",
            "EntryType": "Oem",
            "Id": "1734",
            "Links": {},
            "Message": "The (installation or configuration) job " "JID_559308065328 is successfully completed.",
            "MessageArgs": ["JID_559308065328"],
            "MessageArgs@odata.count": 1,
            "MessageId": "JCP037",
            "Name": "Log Entry 1734",
            "Oem": {
                "Dell": {
                    "@odata.type": "#DellLCLogEntry.v1_0_0.DellLCLogEntry",
                    "Category": "Configuration",
                    "Comment": None,
                    "LastUpdatedByUser": None,
                }
            },
            "OemRecordFormat": "Dell",
            "Severity": "OK",
        },
        {
            "@odata.id": "/redfish/v1/Managers/Testing_oob.1/LogServices/Lclog/Entries/1692",
            "@odata.type": "#LogEntry.v1_6_1.LogEntry",
            "Created": "2022-06-22T16:42:55-05:00",
            "Description": "Log Entry 1692",
            "EntryType": "Oem",
            "Id": "1692",
            "Links": {},
            "Message": "The iDRAC firmware was rebooted with the " "following reason: user initiated.",
            "MessageArgs": ["user initiated"],
            "MessageArgs@odata.count": 1,
            "MessageId": "REBOOT_MSG_ID",
            "Name": "Log Entry 1692",
            "Oem": {
                "Dell": {
                    "@odata.type": "#DellLCLogEntry.v1_0_0.DellLCLogEntry",
                    "Category": "Audit",
                    "Comment": None,
                    "LastUpdatedByUser": None,
                }
            },
            "OemRecordFormat": "Dell",
            "Severity": "OK",
        },
    ],
    "Members@odata.count": 1735,
    "Members@odata.nextLink": "/redfish/v1/Managers/Testing_oob.1/LogServices/Lclog/Entries?$skip=50",
    "Name": "Log Entry Collection",
}

LCLOG_RESPONSE_NO_MESSAGE = deepcopy(LCLOG_RESPONSE)
LCLOG_RESPONSE_NO_MESSAGE["Members"] = []
# The below is trimmed output
SYSTEM_MANAGER_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#ComputerSystem.ComputerSystem",
    "@odata.id": "/redfish/v1/Systems/System.Embedded.1",
    "@odata.type": "#ComputerSystem.v1_16_0.ComputerSystem",
    "Actions": {
        "#ComputerSystem.Reset": {
            "ResetType@Redfish.AllowableValues": [
                "On",
                "ForceOff",
                "ForceRestart",
                "GracefulRestart",
                "GracefulShutdown",
                "PushPowerButton",
                "Nmi",
                "PowerCycle",
            ],
            "target": "/redfish/v1/Systems/System.Embedded.1/Actions/ComputerSystem.Reset",
        }
    },
    "AssetTag": "",
    "Bios": {"@odata.id": "/redfish/v1/Systems/System.Embedded.1/Bios"},
    "BiosVersion": "2.15.1",
    "EthernetInterfaces": {"@odata.id": "/redfish/v1/Systems/System.Embedded.1/EthernetInterfaces"},
    "HostName": "",
    "HostWatchdogTimer": {"FunctionEnabled": False, "Status": {"State": "Disabled"}, "TimeoutAction": "None"},
    "HostingRoles": [],
    "HostingRoles@odata.count": 0,
    "Id": "System.Embedded.1",
    "IndicatorLED": "Lit",
    "LastResetTime": "2023-01-23T09:35:39-06:00",
    "Manufacturer": "Dell Inc.",
    "Model": "PowerEdge R440",
    "Name": "System",
    "NetworkInterfaces": {"@odata.id": "/redfish/v1/Systems/System.Embedded.1/NetworkInterfaces"},
    "PowerState": "On",
    "Status": {"Health": "OK", "HealthRollup": "OK", "State": "Enabled"},
    "SystemType": "Physical",
    "UUID": "4c4c4544-0058-3810-8032-b2c04f525032",
}
UPDATE_SERVICE_RESPONSE = {
    "@odata.context": "/redfish/v1/$metadata#UpdateService.UpdateService",
    "@odata.id": "/redfish/v1/UpdateService",
    "@odata.type": "#UpdateService.v1_11_0.UpdateService",
    "FirmwareInventory": {"@odata.id": "/redfish/v1/UpdateService/FirmwareInventory"},
    "HttpPushUri": "/redfish/v1/UpdateService/FirmwareInventory",
    "Id": "UpdateService",
    "MaxImageSizeBytes": None,
    "MultipartHttpPushUri": "/redfish/v1/UpdateService/MultipartUpload",
    "Name": "Update Service",
    "ServiceEnabled": True,
    "SoftwareInventory": {"@odata.id": "/redfish/v1/UpdateService/SoftwareInventory"},
    "Status": {"Health": "OK", "State": "Enabled"},
}
UPDATE_SERVICE_RESPONSE_NO_HTTP_PUSH = deepcopy(UPDATE_SERVICE_RESPONSE)
del UPDATE_SERVICE_RESPONSE_NO_HTTP_PUSH["HttpPushUri"]


def add_accounts_mock_responses(requests_mock):
    """Setup requests mock URLs and return payloads for all the existing users."""
    requests_mock.get("/redfish/v1/Managers/Testing_oob.1/Accounts", json=ACCOUNTS_RESPONSE)
    users = {"1": "user", "2": "root", "3": "guest"}
    for user_id, username in users.items():
        response = deepcopy(ACCOUNT_RESPONSE)
        response["Id"] = response["Id"].format(user_id=user_id)
        response["@odata.id"] = response["@odata.id"].format(user_id=user_id)
        response["UserName"] = username
        requests_mock.get(
            f"/redfish/v1/AccountService/Accounts/{user_id}", json=response, headers={"ETag": f"12345-{user_id}"}
        )


class RedfishTest(redfish.Redfish):
    """An inherited class used for testing."""

    system = "Testing_system.1"
    manager = "Testing_oob.1"
    log_service = "Testing_oob.1"
    reboot_message_id = "REBOOT_MSG_ID"

    def get_power_state(self) -> str:
        """Return the current power state of the device."""
        return "On"


class TestRedfish:
    """Tests for the Redfish class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        interface = ipaddress.ip_interface("10.0.0.1/16")
        self.redfish = RedfishTest("test01", interface, "root", "mysecret", dry_run=False)
        self.redfish_dry_run = RedfishTest("test01", interface, "root", "mysecret", dry_run=True)
        self.requests_mock = requests_mock

    def test_property_magic_str(self):
        """It should equal the fqdn."""
        assert str(self.redfish) == "root@test01 (10.0.0.1)"

    def test_property_hostname(self):
        """It should equal the fqdn."""
        assert self.redfish.hostname == "test01"

    def test_property_interface(self):
        """It should equal the fqdn."""
        assert isinstance(self.redfish.interface, ipaddress.IPv4Interface)
        assert str(self.redfish.interface.ip) == "10.0.0.1"

    @pytest.mark.parametrize(
        "response, reboot_time",
        ((LCLOG_RESPONSE, "2022-06-22T16:42:55-05:00"), (LCLOG_RESPONSE_NO_MESSAGE, "1970-01-01T00:00:00-00:00")),
    )
    def test_last_reboot(self, response, reboot_time):
        """Return the last reboot time."""
        reboot_time = datetime.fromisoformat(reboot_time)
        self.requests_mock.get(self.redfish.log_entries, json=response)
        assert self.redfish.last_reboot() == reboot_time

    def test_wait_reboot_since(self):
        """It should raise an error if the reboot time is to early."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        self.requests_mock.get(self.redfish.log_entries, json=LCLOG_RESPONSE)
        since = datetime.fromisoformat("2022-01-01T00:05:00-00:00")
        self.redfish.wait_reboot_since(since)

    @mock.patch("spicerack.redfish.time.sleep")
    def test_wait_reboot_since_to_early(self, _mocked_sleep):
        """It should raise an error if the reboot time is to early."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        self.requests_mock.get(self.redfish.log_entries, json=LCLOG_RESPONSE_NO_MESSAGE)
        since = datetime.fromisoformat("2022-01-01T00:05:00-00:00")
        with pytest.raises(redfish.RedfishError, match="no new reboot detected"):
            self.redfish.wait_reboot_since(since)

    def test_property_system_info(self):
        """It should return the firmware."""
        self.requests_mock.get(self.redfish.system_manager, json=SYSTEM_MANAGER_RESPONSE)
        assert self.redfish.system_info == SYSTEM_MANAGER_RESPONSE
        # try again to hit the cached version
        assert self.redfish.system_info == SYSTEM_MANAGER_RESPONSE

    def test_property_updateservice_info(self):
        """It should return the firmware."""
        self.requests_mock.get(self.redfish.update_service, json=UPDATE_SERVICE_RESPONSE)
        assert self.redfish.updateservice_info == UPDATE_SERVICE_RESPONSE
        # try again to hit the cached version
        assert self.redfish.updateservice_info == UPDATE_SERVICE_RESPONSE

    def test_property_multipushuri(self):
        """It should return the multipushuri."""
        self.requests_mock.get(self.redfish.update_service, json=UPDATE_SERVICE_RESPONSE)
        assert self.redfish.multipushuri == "/redfish/v1/UpdateService/MultipartUpload"

    def test_property_bios(self):
        """It should return the firmware."""
        self.requests_mock.get(self.redfish.system_manager, json=SYSTEM_MANAGER_RESPONSE)
        assert self.redfish.bios_version == version.parse("2.15.1")

    def test_property_model(self):
        """It should return the firmware."""
        self.requests_mock.get(self.redfish.system_manager, json=SYSTEM_MANAGER_RESPONSE)
        assert self.redfish.model == "PowerEdge R440"

    def test_property_manufacturer(self):
        """It should return the firmware."""
        self.requests_mock.get(self.redfish.system_manager, json=SYSTEM_MANAGER_RESPONSE)
        assert self.redfish.manufacturer == "Dell Inc."

    def test_property_firmware(self):
        """It should return the firmware."""
        self.requests_mock.get(self.redfish.oob_manager, json=MANAGER_RESPONSE)
        assert self.redfish.firmware_version == version.parse("6.00.30.00")
        # assert twice to check cached version
        assert self.redfish.firmware_version == version.parse("6.00.30.00")

    def test_property_oob_model(self):
        """It should return the model."""
        self.requests_mock.get(self.redfish.oob_manager, json=MANAGER_RESPONSE)
        assert self.redfish.oob_model == "14G Monolithic"
        # assert twice to check cached version
        assert self.redfish.oob_model == "14G Monolithic"

    def test_property_missing_pushuri(self):
        """It should return the pushuri."""
        self.requests_mock.get(self.redfish.update_service, json=UPDATE_SERVICE_RESPONSE_NO_HTTP_PUSH)
        with pytest.raises(NotImplementedError):
            assert self.redfish.pushuri is None

    def test_most_recent_member(self):
        """It should return the item with the most recent date."""
        data = [
            {"date": "2022-01-01T00:00:00-00:00"},
            {"date": "1970-01-01T00:00:00-00:00"},
            {"date": "1971-01-01T00:00:00-00:00"},
        ]
        assert self.redfish.most_recent_member(data, "date") == {"date": "2022-01-01T00:00:00-00:00"}

    @pytest.mark.parametrize("method", ("get", "head"))
    def test_request_dry_run_ro(self, method):
        """It should perform any RO request and return the actual response also in dry_run mode."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        self.requests_mock.head("/redfish")
        response = self.redfish_dry_run.request(method, "/redfish")
        assert response.status_code == 200
        if method == "get":
            assert response.json() == {"v1": "/redfish/v1/"}

    @pytest.mark.parametrize("method", ("connect", "delete", "options", "patch", "post", "put", "trace"))
    def test_request_dry_run_rw(self, method):
        """It should not perform any RW request and return a dummy successful response in dry-run mode."""
        response = self.redfish_dry_run.request(method, "/redfish")
        assert response.status_code == 200
        assert not response.text

    def test_request_dry_run_fail(self, caplog):
        """If the request fails in dry-run mode, it should return a dummy successful response."""
        self.requests_mock.get("/redfish", exc=requests.exceptions.ConnectTimeout)
        with caplog.at_level(logging.ERROR):
            response = self.redfish_dry_run.request("get", "/redfish")

        assert response.status_code == 200
        assert "Failed to perform GET request to https://10.0.0.1/redfish" in caplog.text

    def test_request_ok(self):
        """It should perform the provided request and return it."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        response = self.redfish.request("get", "/redfish")
        assert response.json() == {"v1": "/redfish/v1/"}
        assert response.status_code == 200

    def test_request_response_wrong_status_code(self):
        """It should raise a RedfishError if the request returns an error status code."""
        self.requests_mock.post("/redfish", json={"error": {"code": "1.0", "message": "error"}}, status_code=405)
        with pytest.raises(redfish.RedfishError, match="POST https://10.0.0.1/redfish returned HTTP 405 with message"):
            self.redfish.request("post", "/redfish", json={"key": "value"})

    def test_request_response_raises(self):
        """It should raise a RedfishError if the request failes to be performed."""
        self.requests_mock.get("/redfish", exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(redfish.RedfishError, match="Failed to perform GET request to https://10.0.0.1/redfish"):
            self.redfish.request("get", "/redfish")

    def test_request_invalid_uri(self):
        """It should raise a RedfishError if the URI is invalid."""
        with pytest.raises(redfish.RedfishError, match="Invalid uri redfish, it must start with a /"):
            self.redfish.request("get", "redfish")

    def test_submit_task_dry_run(self):
        """In dry-run mode should not submit a task and return a dummy location."""
        assert self.redfish_dry_run.submit_task("/redfish/v1/SomeAction") == "/"

    def test_submit_task_ok(self):
        """It should submit the request and return the URI for polling the task results."""
        self.requests_mock.post(
            "/redfish/v1/SomeAction",
            status_code=202,
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
        )
        response = self.redfish.submit_task("/redfish/v1/SomeAction")
        assert response == "/redfish/v1/TaskService/Tasks/JID_1234567890"

    def test_submit_task_invalid_code(self):
        """It should raise a RedfishError if the status code is not 202."""
        self.requests_mock.post(
            "/redfish/v1/SomeAction",
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
        )
        with pytest.raises(redfish.RedfishError, match="expected HTTP 202, got HTTP 200 instead"):
            self.redfish.submit_task("/redfish/v1/SomeAction")

    def test_submit_task_no_location(self):
        """It should raise a RedfishError if there is no Location header in the response."""
        self.requests_mock.post("/redfish/v1/SomeAction", status_code=202)
        with pytest.raises(redfish.RedfishError, match="Unable to get the task URI to poll results"):
            self.redfish.submit_task("/redfish/v1/SomeAction")

    def test_connection_ok(self):
        """It should not raise if able to connect to the Redfish API."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        self.redfish.check_connection()

    def test_submit_task_file(self):
        """It should submit the request and return the URI for polling the task results."""
        self.requests_mock.get(self.redfish.update_service, json=UPDATE_SERVICE_RESPONSE)
        self.requests_mock.post(
            self.redfish.multipushuri,
            status_code=202,
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
        )
        assert self.redfish.submit_files({}) == "/redfish/v1/TaskService/Tasks/JID_1234567890"

    def test_submit_file_dry_run(self):
        """In dry-run mode should not submit a task and return a dummy location."""
        assert self.redfish_dry_run.submit_files({}) == "/"

    @mock.patch("spicerack.redfish.Path.open")
    @mock.patch("spicerack.redfish.Redfish.submit_files")
    def test_submit_upload_file(self, mock_submit_files, mocked_path_open):
        """In dry-run mode should not submit a task and return a dummy location."""
        mock_submit_files.return_value = "/redfish/v1/TaskService/Tasks/JID_1234567890"
        assert self.redfish.upload_file(Path("/foo/test")) == "/redfish/v1/TaskService/Tasks/JID_1234567890"
        mocked_path_open.assert_called_once_with("rb")

    @pytest.mark.parametrize("reboot", (True, False))
    @mock.patch("spicerack.redfish.Redfish.submit_files")
    def test_submit_multipush_upload(self, mock_submit_files, reboot):
        """In dry-run mode should not submit a task and return a dummy location."""
        mock_submit_files.return_value = "/redfish/v1/TaskService/Tasks/JID_1234567890"
        data = BytesIO()
        assert self.redfish.multipush_upload("test", data, reboot) == "/redfish/v1/TaskService/Tasks/JID_1234567890"

    def test_connection_fail(self):
        """It should raise a RedfishError if unable to connect to the Redfish API."""
        self.requests_mock.get("/redfish", status_code=400)
        with pytest.raises(redfish.RedfishError, match="GET https://10.0.0.1/redfish returned HTTP 400"):
            self.redfish.check_connection()

    def test_poll_task_dry_run(self):
        """It should return a dummy response in dry-run mode."""
        assert self.redfish_dry_run.poll_task("/redfish/v1/TaskService/Tasks/JID_1234567890") == {}

    @pytest.mark.parametrize("response", (DELL_TASK_REPONSE, DELL_TASK_REPONSE_EPOC, DELL_TASK_REPONSE_BAD_TIME))
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_poll_endtime(self, mocked_sleep, response):
        """It should raise a RedfishError if polling the task the device returns a failure code."""
        self.requests_mock.get("/redfish/v1/TaskService/Tasks/JID_1234567890", status_code=202, json=response)
        with pytest.raises(redfish.RedfishTaskNotCompletedError):
            self.redfish.poll_task("/redfish/v1/TaskService/Tasks/JID_1234567890")
        mocked_sleep.assert_called()

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_poll_task_raises(self, mocked_sleep):
        """It should raise a RedfishError if polling the task the device returns a failure code."""
        self.requests_mock.get("/redfish/v1/TaskService/Tasks/JID_1234567890", status_code=201)
        with pytest.raises(redfish.RedfishError, match="JID_1234567890 returned HTTP 201"):
            self.redfish.poll_task("/redfish/v1/TaskService/Tasks/JID_1234567890")

        mocked_sleep.assert_not_called()

    def test_find_account_ok(self):
        """It should return the URI of the account with the given username."""
        add_accounts_mock_responses(self.requests_mock)
        uri, etag = self.redfish.find_account("root")
        assert uri == "/redfish/v1/AccountService/Accounts/2"
        assert etag == "12345-2"

    def test_find_account_raises(self):
        """It should raise a RedfishError if the user is not found."""
        add_accounts_mock_responses(self.requests_mock)
        with pytest.raises(redfish.RedfishError, match="Unable to find account for username nonexistent"):
            self.redfish.find_account("nonexistent")

    @pytest.mark.parametrize("user_id, username, is_current", ((1, "user", False), (2, "root", True)))
    def test_change_user_password_ok(self, user_id, username, is_current):
        """It should change the password for the given user and update the instance auth credentials accordingly."""
        self.requests_mock.get("/redfish", json={"v1": "/redfish/v1/"})
        add_accounts_mock_responses(self.requests_mock)
        response = {
            "@Message.ExtendedInfo": [
                {"Message": "Changed password", "MessageId": "123", "Severity": "Informational", "Resolution": "None"}
            ]
        }
        self.requests_mock.patch(f"/redfish/v1/AccountService/Accounts/{user_id}", json=response)

        current_auth = self.redfish.request("get", "/redfish").request.headers["Authorization"]
        self.redfish.change_user_password(username, "test1234")
        new_auth = self.redfish.request("get", "/redfish").request.headers["Authorization"]

        if is_current:  # Check that the authorization header changed
            assert new_auth != current_auth
        else:
            assert new_auth == current_auth

    def test_change_user_password_raises(self):
        """It should raise a RedfishError if the reponse is not what HTTP 200."""
        add_accounts_mock_responses(self.requests_mock)
        self.requests_mock.patch("/redfish/v1/AccountService/Accounts/2", status_code=202)
        with pytest.raises(redfish.RedfishError, match="Got unexpected HTTP 202, expected 200"):
            self.redfish.change_user_password("root", "test1234")

    @pytest.mark.parametrize("action", tuple(redfish.ChassisResetPolicy))
    def test_chassis_reset_ok(self, action):
        """It should perform a chassis reset with the given action."""
        self.requests_mock.post("/redfish/v1/Systems/Testing_system.1/Actions/ComputerSystem.Reset", status_code=204)
        self.redfish.chassis_reset(action)

    def test_chassis_reset_raises(self):
        """It should raise a RedfishError if the response code of the chassis reset operation is not 204."""
        self.requests_mock.post("/redfish/v1/Systems/Testing_system.1/Actions/ComputerSystem.Reset")
        with pytest.raises(redfish.RedfishError, match="Got unexpected response HTTP 200, expected HTTP 204"):
            self.redfish.chassis_reset(redfish.ChassisResetPolicy.FORCE_OFF)


class TestDellSCP:
    """Tests for the DellSCP class."""

    def setup_method(self):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.config = redfish.DellSCP(deepcopy(DELL_SCP), redfish.DellSCPTargetPolicy.ALL)
        self.config_allow_new = redfish.DellSCP(
            deepcopy(DELL_SCP), redfish.DellSCPTargetPolicy.ALL, allow_new_attributes=True
        )

    @pytest.mark.parametrize(
        "property_name, expected",
        (
            ("config", DELL_SCP),
            ("target", redfish.DellSCPTargetPolicy.ALL),
            ("service_tag", "12ABC34"),
            ("model", "PowerEdge R440"),
            ("timestamp", datetime(2021, 12, 9, 9, 32, 6)),
            ("comments", ["First comment"]),
            (
                "components",
                {
                    "Some.Component.1": {"Some.Attribute.1": "value"},
                    "Some.Component.2": {"Some.Attribute.1": "value", "Some.Attribute.2": "value"},
                    "List.Component.1": {
                        "Comma.Separated.List.1": "value1,value2",
                        "Comma.Space.Separated.List.1": "value1, value2",
                    },
                },
            ),
        ),
    )
    def test_properties(self, property_name, expected):
        """It should return the value of the given property."""
        assert getattr(self.config, property_name) == expected
        assert getattr(self.config_allow_new, property_name) == expected

    @pytest.mark.parametrize(
        "component, attribute, exception_message",
        (
            ("Non.Existent", "Some.Attribute.1", "Unable to find component Non.Existent"),
            ("Some.Component.1", "Non.Existent", "Unable to find attribute Some.Component.1 -> Non.Existent"),
        ),
    )
    def test_set_raise(self, component, attribute, exception_message):
        """It should raise RedfishError if the specified component or attribute does not exists and is now allowed."""
        with pytest.raises(redfish.RedfishError, match=exception_message):
            self.config.set(component, attribute, "new_value")

    def test_set_new_attribute(self):
        """It should add the new attribute if allow_new_attributes is set to True or empty_components() was called."""
        params = ["Some.Component.1", "Non.Existent", "new_value"]
        self.config_allow_new.set(*params)
        assert self.config_allow_new.components[params[0]][params[1]] == params[2]
        self.config.empty_components()
        self.config.set(*params)
        assert self.config.components[params[0]][params[1]] == params[2]

    def test_set_raise_new_component(self):
        """It should raise RedfishError if the component does not exists even if allow_new_attributes is set."""
        with pytest.raises(redfish.RedfishError, match="Unable to find component Non.Existent"):
            self.config_allow_new.set("Non.Existent", "Some.Attribute.1", "new_value")

    def test_set_new_component(self):
        """It should create a new component and add to it the new attribute if empty_components() was called."""
        self.config.empty_components()
        self.config.set("Non.Existent", "Some.Attribute.1", "new_value")
        assert self.config.components["Non.Existent"]["Some.Attribute.1"] == "new_value"

    @pytest.mark.parametrize(
        "index, component, attribute, value",
        (
            (1, "Some.Component.1", "Some.Attribute.1", "value"),
            (2, "List.Component.1", "Comma.Separated.List.1", "value1,value2"),
            (2, "List.Component.1", "Comma.Space.Separated.List.1", "value1, value2"),
        ),
    )
    def test_set_same_value(self, index, component, attribute, value, caplog):
        """It should not set the same value and log a different message if the value is already correct."""
        with caplog.at_level(logging.INFO):
            was_changed = self.config.set(component, attribute, value)

        assert not was_changed
        assert f"Skipped set of attribute {component} -> {attribute}, has already" in caplog.text
        assert all(
            i["Set On Import"] == "False"
            for i in self.config.config["SystemConfiguration"]["Components"][index]["Attributes"]
        )

    def test_set_new_value(self):
        """It should set the value and mark it for import."""
        was_changed = self.config.set("Some.Component.1", "Some.Attribute.1", "new value")
        attribute = self.config.config["SystemConfiguration"]["Components"][1]["Attributes"][0]
        assert was_changed
        assert attribute["Value"] == "new value"
        assert attribute["Set On Import"] == "True"

    def test_update(self):
        """It should update the config with the given changes."""
        changes = {
            "Some.Component.1": {"Some.Attribute.1": "value"},
            "Some.Component.2": {"Some.Attribute.2": "new value"},
        }
        was_changed = self.config.update(changes)
        assert was_changed
        assert self.config.components["Some.Component.1"]["Some.Attribute.1"] == "value"
        assert self.config.components["Some.Component.2"]["Some.Attribute.2"] == "new value"

    def test_update_no_changes(self):
        """It should return False if no changes were made."""
        changes = {
            "Some.Component.1": {"Some.Attribute.1": "value"},
            "Some.Component.2": {"Some.Attribute.2": "value"},
        }
        was_changed = self.config.update(changes)
        assert not was_changed
        assert self.config.components["Some.Component.1"]["Some.Attribute.1"] == "value"
        assert self.config.components["Some.Component.2"]["Some.Attribute.2"] == "value"

    def test_empty_components(self):
        """It should empty the components of the current configuration."""
        self.config.empty_components()
        assert self.config.components == {}


class TestRedfishDell:
    """Tests for the RedfishDell class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        interface = ipaddress.ip_interface("10.0.0.1/16")
        self.redfish = redfish.RedfishDell("test01", interface, "root", "mysecret", dry_run=False)
        self.requests_mock = requests_mock

    def test_property_system_manager(self):
        """It should return the oob_manager."""
        assert self.redfish.system_manager == "/redfish/v1/Systems/System.Embedded.1"

    def test_property_oob_manager(self):
        """It should return the oob_manager."""
        assert self.redfish.oob_manager == "/redfish/v1/Managers/iDRAC.Embedded.1"

    def test_property_storage_manager(self):
        """It should return the storage_manager."""
        assert self.redfish.storage_manager == "/redfish/v1/Systems/System.Embedded.1/Storage"

    @pytest.mark.parametrize(
        "response, endpoint",
        (
            (MANAGER_RESPONSE, "/redfish/v1/Managers/iDRAC.Embedded.1/LogServices/Lclog/Entries"),
            (MANAGER_RESPONSE_V3, "/redfish/v1/Managers/Logs/Lclog"),
        ),
    )
    def test_property_log_entries(self, response, endpoint) -> str:
        """String representing the uri for the log entries."""
        self.requests_mock.get(self.redfish.oob_manager, json=response)
        assert self.redfish.log_entries == endpoint

    def test_property_reboot_message_id(self) -> str:
        """Property to return the Message Id for reboot log entries."""
        assert self.redfish.reboot_message_id == "RAC0182"

    @pytest.mark.parametrize("response, generation", ((MANAGER_RESPONSE, 14), (MANAGER_RESPONSE_BAD, 1)))
    def test_property_generation(self, response, generation):
        """It should return the generation."""
        self.requests_mock.get(self.redfish.oob_manager, json=response)
        assert self.redfish.generation == generation
        # assert twice to check cached version
        assert self.redfish.generation == generation

    @pytest.mark.parametrize("generation", (1, 13, 14))
    @mock.patch("spicerack.redfish.time.sleep")
    @mock.patch("spicerack.redfish.RedfishDell.last_reboot")
    @mock.patch("spicerack.redfish.RedfishDell.check_connection")
    def test_wait_reboot_since(self, mocked_check_connection, mocked_last_reboot, mocked_sleep, generation):
        """It should return immediately if the host has already rebooted."""
        since = datetime.fromisoformat("2022-01-01T00:00:00-00:00")
        mocked_check_connection.return_value = True
        mocked_last_reboot.return_value = datetime.fromisoformat("2022-01-01T00:05:00-00:00")
        self.redfish._generation = generation  # pylint: disable=protected-access
        self.redfish.wait_reboot_since(since)
        mocked_check_connection.assert_called_once()
        mocked_last_reboot.assert_called_once_with()
        calls = []
        if generation < 14:
            calls.append(mock.call(120))
        calls.append(mock.call(30))
        assert mocked_sleep.mock_calls == calls

    @pytest.mark.parametrize("generation", (1, 13, 14))
    @mock.patch("spicerack.redfish.time.sleep")
    @mock.patch("spicerack.redfish.RedfishDell.last_reboot")
    @mock.patch("spicerack.redfish.RedfishDell.check_connection")
    def test_wait_reboot_since_to_early(self, mocked_check_connection, mocked_last_reboot, mocked_sleep, generation):
        """It should raise an error if the reboot time is to early."""
        since = datetime.fromisoformat("2022-01-01T00:05:00-00:00")
        mocked_check_connection.return_value = True
        mocked_last_reboot.return_value = datetime.fromisoformat("2022-01-01T00:00:00-00:00")
        self.redfish._generation = generation  # pylint: disable=protected-access
        with pytest.raises(redfish.RedfishError, match="no new reboot detected"):
            self.redfish.wait_reboot_since(since)
        mocked_check_connection.assert_called_with()
        assert mocked_check_connection.call_count == 240
        mocked_last_reboot.assert_called_with()
        assert mocked_last_reboot.call_count == 240
        calls = []
        if generation < 14:
            calls = [mock.call(120), mock.call(10)] * 239 + [mock.call(120)]
        else:
            calls = [mock.call(10)] * 239

        assert mocked_sleep.mock_calls == calls

    @pytest.mark.parametrize("allow_new", (False, True))
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_scp_dump(self, mocked_sleep, allow_new):
        """It should return an instance of DellSCP with the current configuration for the given target."""
        self.requests_mock.post(
            "/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager.ExportSystemConfiguration",
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
            status_code=202,
        )
        self.requests_mock.get(
            "/redfish/v1/TaskService/Tasks/JID_1234567890",
            [{"status_code": 202, "json": DELL_TASK_REPONSE}, {"status_code": 200, "json": DELL_SCP}],
        )
        config = self.redfish.scp_dump(allow_new_attributes=allow_new)
        assert config.service_tag == "12ABC34"
        assert mocked_sleep.called
        if allow_new:
            config.set("Some.Component.1", "Non.Existent", "new_value")
            assert config.components["Some.Component.1"]["Non.Existent"] == "new_value"
        else:
            with pytest.raises(redfish.RedfishError, match="Unable to find attribute Some.Component.1 -> Non.Existent"):
                config.set("Some.Component.1", "Non.Existent", "new_value")

    @pytest.mark.parametrize(
        "uri_suffix, preview",
        (
            ("ImportSystemConfigurationPreview", True),
            ("ImportSystemConfiguration", False),
        ),
    )
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_scp_push(self, mocked_sleep, uri_suffix, preview):
        """It should push the configuration to the device for preview, no changes will be applied."""
        expected = deepcopy(DELL_TASK_REPONSE)
        expected["EndTime"] = "2021-12-09T14:39:29-06:00"
        self.requests_mock.post(
            f"/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager.{uri_suffix}",
            headers={"Location": "/redfish/v1/TaskService/Tasks/JID_1234567890"},
            status_code=202,
        )
        self.requests_mock.get(
            "/redfish/v1/TaskService/Tasks/JID_1234567890",
            [{"status_code": 202, "json": DELL_TASK_REPONSE}, {"status_code": 200, "json": expected}],
        )
        result = self.redfish.scp_push(redfish.DellSCP(DELL_SCP, redfish.DellSCPTargetPolicy.ALL), preview=preview)
        assert result == expected
        assert mocked_sleep.called

    def test_get_power_state(self):
        """It should return the current power state of the device."""
        self.requests_mock.get("/redfish/v1/Chassis/System.Embedded.1", json={"PowerState": "On"})
        assert self.redfish.get_power_state() == "On"


class TestRedfishSupermicro:
    """Tests for the RedfishSupermicro class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        interface = ipaddress.ip_interface("10.0.0.1/16")
        self.redfish = redfish.RedfishSupermicro("test01", interface, "root", "mysecret", dry_run=False)
        self.requests_mock = requests_mock

    def test_property_system_manager(self):
        """It should return the oob_manager."""
        assert self.redfish.system_manager == "/redfish/v1/Systems/1"

    def test_property_oob_manager(self):
        """It should return the oob_manager."""
        assert self.redfish.oob_manager == "/redfish/v1/Managers/1"

    def test_property_storage_manager(self):
        """It should return the storage_manager."""
        assert self.redfish.storage_manager == "/redfish/v1/Systems/1/Storage"

    def test_property_log_entries(self) -> str:
        """String representing the uri for the log entries."""
        assert self.redfish.log_entries == "/redfish/v1/Managers/1/LogServices/Log1/Entries"

    def test_property_reboot_message_id(self) -> str:
        """Property to return the Message Id for reboot log entries."""
        assert self.redfish.reboot_message_id == "Event.1.0.SystemPowerAction"

    def test_get_power_state(self):
        """It should return the current power state of the device."""
        self.requests_mock.get("/redfish/v1/Systems/1", json={"PowerState": "On"})
        assert self.redfish.get_power_state() == "On"



"""DHCP Module Tests."""
import base64
import re
from hashlib import sha256
from ipaddress import IPv4Address
from unittest import mock

import pytest

from spicerack import dhcp
from spicerack.locking import NoLock
from spicerack.remote import RemoteExecutionError


def get_mock_hosts():
    """Return a `spicerack.remote.Hosts` mock."""
    hosts = mock.MagicMock()
    hosts.__len__.return_value = 1
    hosts.run_sync.return_value = "some value"
    return hosts


def get_mock_fail_hosts():
    """Return a `spicerack.remote.Hosts` mock where execution fails."""
    hosts = get_mock_hosts()
    hosts.run_sync.side_effect = RemoteExecutionError("mock error", 1, iter(()))
    return hosts


def get_mock_suc_fail_hosts():
    """Return a `spicerack.remote.Hosts` mock where execution succeeds and then fails."""
    hosts = get_mock_hosts()
    hosts.run_sync.side_effect = [iter(()), RemoteExecutionError("mock error", 1, iter(()))]
    return hosts


def get_mock_config():
    """Return a `spicerack.dhcp.Configuration` mock."""
    config = mock.MagicMock()
    config.__str__.return_value = "test configuration"
    config.config_base64 = base64.b64encode(b"test configuration").decode()
    config.filename = "test.conf"
    return config


# Test Configuration Generator Objects
configuration_generator_data = (
    # dhcpconfopt82 tests
    # - basic check of functionality
    (
        dhcp.DHCPConfOpt82,
        {
            "hostname": "testhost0",
            "ipv4": IPv4Address("10.0.0.1"),
            "switch_hostname": "asw2-d-eqiad",
            "switch_iface": "ge-0/0/0",
            "vlan": 1021,
            "ttys": 1,
            "distro": "buster",
        },
        (
            "\nhost testhost0 {\n"
            '    host-identifier option agent.circuit-id "asw2-d-eqiad:ge-0/0/0:1021";\n'
            "    fixed-address 10.0.0.1;\n"
            '    option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/buster-installer/";\n'
            "}\n"
        ),
        "ttyS1-115200/testhost0.conf",
    ),
    # - update media type
    (
        dhcp.DHCPConfOpt82,
        {
            "hostname": "testhost0",
            "ipv4": IPv4Address("10.0.0.1"),
            "switch_hostname": "asw2-d-eqiad",
            "switch_iface": "ge-0/0/0",
            "vlan": 1021,
            "ttys": 1,
            "distro": "buster",
            "media_type": "rescue",
        },
        (
            "\nhost testhost0 {\n"
            '    host-identifier option agent.circuit-id "asw2-d-eqiad:ge-0/0/0:1021";\n'
            "    fixed-address 10.0.0.1;\n"
            '    option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/buster-rescue/";\n'
            "}\n"
        ),
        "ttyS1-115200/testhost0.conf",
    ),
    # - tty argument should change the file path
    (
        dhcp.DHCPConfOpt82,
        {
            "hostname": "testhost0",
            "ipv4": IPv4Address("10.0.0.1"),
            "switch_hostname": "asw2-d-eqiad",
            "switch_iface": "ge-0/0/0",
            "vlan": 1021,
            "ttys": 0,
            "distro": "buster",
        },
        (
            "\nhost testhost0 {\n"
            '    host-identifier option agent.circuit-id "asw2-d-eqiad:ge-0/0/0:1021";\n'
            "    fixed-address 10.0.0.1;\n"
            '    option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/buster-installer/";\n'
            "}\n"
        ),
        "ttyS0-115200/testhost0.conf",
    ),
    # DHCPConfMac tests
    # - basic check of functionality
    (
        dhcp.DHCPConfMac,
        {
            "hostname": "testhost0",
            "ipv4": IPv4Address("10.0.0.1"),
            "mac": "00:00:00:00:00:01",
            "ttys": 0,
            "distro": "buster",
        },
        (
            "\nhost testhost0 {\n"
            "    hardware ethernet 00:00:00:00:00:01;\n"
            "    fixed-address 10.0.0.1;\n"
            '    option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/buster-installer/";\n'
            "}\n"
        ),
        "ttyS0-115200/testhost0.conf",
    ),
    # - update media type
    (
        dhcp.DHCPConfMac,
        {
            "hostname": "testhost0",
            "ipv4": IPv4Address("10.0.0.1"),
            "mac": "00:00:00:00:00:01",
            "ttys": 0,
            "distro": "buster",
            "media_type": "installer-11.0",
        },
        (
            "\nhost testhost0 {\n"
            "    hardware ethernet 00:00:00:00:00:01;\n"
            "    fixed-address 10.0.0.1;\n"
            '    option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/buster-installer-11.0/";\n'
            "}\n"
        ),
        "ttyS0-115200/testhost0.conf",
    ),
    # - tty argument should change the file path
    (
        dhcp.DHCPConfMac,
        {
            "hostname": "testhost0",
            "ipv4": IPv4Address("10.0.0.1"),
            "mac": "00:00:00:00:00:01",
            "ttys": 1,
            "distro": "bullseye",
        },
        (
            "\nhost testhost0 {\n"
            "    hardware ethernet 00:00:00:00:00:01;\n"
            "    fixed-address 10.0.0.1;\n"
            '    option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/bullseye-installer/";\n'
            "}\n"
        ),
        "ttyS1-115200/testhost0.conf",
    ),
    # dhcpconfmgmt tests
    (  # Dell host
        dhcp.DHCPConfMgmt,
        {
            "datacenter": "eqiad",
            "serial": "TEST",
            "manufacturer": "Dell",
            "fqdn": "test1001.mgmt.eqiad.wmnet",
            "ipv4": IPv4Address("10.0.0.1"),
        },
        (
            '\nclass "test1001.mgmt.eqiad.wmnet" {\n'
            '    match if (lcase(option host-name) = "idrac-test");\n'
            "}\npool {\n"
            '    allow members of "test1001.mgmt.eqiad.wmnet";\n'
            "    range 10.0.0.1 10.0.0.1;\n"
            "}\n"
        ),
        "mgmt-eqiad/test1001.mgmt.eqiad.wmnet.conf",
    ),
    (  # Juniper device
        dhcp.DHCPConfMgmt,
        {
            "datacenter": "eqiad",
            "serial": "TESTSERIAL",
            "manufacturer": "Juniper",
            "fqdn": "test-switch1001.mgmt.eqiad.wmnet",
            "ipv4": IPv4Address("10.0.0.1"),
        },
        (
            '\nclass "test-switch1001.mgmt.eqiad.wmnet" {\n'
            '    match if (lcase(option host-name) = "testserial");\n'
            "}\npool {\n"
            '    allow members of "test-switch1001.mgmt.eqiad.wmnet";\n'
            "    range 10.0.0.1 10.0.0.1;\n"
            "}\n"
        ),
        "mgmt-eqiad/test-switch1001.mgmt.eqiad.wmnet.conf",
    ),
)
"""`tuple[class, tuple[dict[str, str], str]]`: Parameters for test_configuration_generator."""


@pytest.mark.parametrize("generator,kw_arguments,expected,expected_filename", configuration_generator_data)
def test_configuration_generator(generator, kw_arguments, expected, expected_filename):
    """Test configuration generators producing expected outputs with various parameters."""
    confobj = generator(**kw_arguments)
    assert str(confobj) == expected
    assert confobj.filename == expected_filename


@pytest.mark.parametrize(
    "mac",
    (
        "00:00:00:00:00:00",
        "01:23:45:67:89:ab",
        "cd:ef:00:11:22:33",
        "aa:aa:aa:aa:aa:aa",
        "AA:AA:AA:AA:AA:AA",
        "ff:ff:ff:ff:ff:ff",
    ),
)
def test_dhcp_conf_mac_valid_mac(mac):
    """It should not raise when a valid MAC address is passed."""
    config = dhcp.DHCPConfMac(hostname="testhost0", ipv4=IPv4Address("10.0.0.1"), mac=mac, ttys=1, distro="bullseye")
    assert config.mac == mac


@pytest.mark.parametrize(
    "mac",
    (
        ":00:00:00:00:00:00",
        "00:00:00:00:00:00:",
        ":00:00:00:00:00:00:",
        "0:00:00:00:00:00",
        "00:00:00:00:00:0",
        "000:00:00:00:00:00",
        "00:00:00:00:00:000",
        "00:00:000:00:00:00",
        "g0:00:00:00:00:00",
        "00-00-00-00-00-00",
        "01-23-45-67-89-AB",
        "0123.4567.89AB",
        "01-23-45-67-89-AB-CD-EF",
    ),
)
def test_dhcp_conf_mac_invalid_mac(mac):
    """It should raise DHCPError if an invalid MAC address is passed."""
    with pytest.raises(dhcp.DHCPError, match="Invalid MAC address"):
        dhcp.DHCPConfMac(hostname="testhost0", ipv4=IPv4Address("10.0.0.1"), mac=mac, ttys=1, distro="bullseye")


@pytest.mark.parametrize(
    "datacenter, fqdn, error",
    (
        ("invalid", "", "Invalid datacenter invalid, must be one of"),
        ("eqiad", "invalid.eqiad.wmnet", "Invalid management FQDN invalid.eqiad.wmnet, must match"),
    ),
)
def test_dhcp_mgmt_fail(datacenter, fqdn, error):
    """A DHCPConfMgmt object should fail to create if invalid parameters are passed to its init."""
    with pytest.raises(dhcp.DHCPError, match=re.escape(error)):
        dhcp.DHCPConfMgmt(datacenter=datacenter, serial="", manufacturer="", fqdn=fqdn, ipv4=None)


def test_create_dhcp_no_hosts():
    """Test fail (hosts parameter has no hosts) DHCP instance creation."""
    hosts_mock = get_mock_hosts()
    hosts_mock.__len__.return_value = 0
    with pytest.raises(dhcp.DHCPError, match="No target hosts provided"):
        dhcp.DHCP(hosts_mock, datacenter="eqiad", lock=NoLock())


def test_create_dhcp_wrong_dc():
    """It should raise a DHCPError in case the datacenter is inexistent."""
    hosts_mock = get_mock_hosts()
    hosts_mock.__len__.return_value = 2
    with pytest.raises(dhcp.DHCPError, match="Invalid datacenter invalid, must be one of"):
        dhcp.DHCP(hosts_mock, datacenter="invalid", lock=NoLock())


class TestDHCP:
    """Test various other aspects of DHCP module."""

    def setup_method(self):
        """Do any one time setup for the tests."""
        # pylint: disable=attribute-defined-outside-init
        remotehosts_mock = get_mock_hosts()
        self.dhcp = dhcp.DHCP(remotehosts_mock, datacenter="eqiad", lock=NoLock(), dry_run=False)

    def _setup_dhcp_mocks(self, hosts=None):
        """Setup the DHCP's hosts remote as new mocks."""
        if hosts is None:
            hosts = get_mock_hosts()

        self.dhcp._hosts = hosts  # pylint: disable=protected-access
        return hosts

    def test_push_configuration_ok(self):
        """Test push_configuration success."""
        config = get_mock_config()
        hosts = self._setup_dhcp_mocks()

        self.dhcp.push_configuration(config)
        hosts.run_sync.assert_has_calls(
            [
                mock.call(
                    "/bin/echo 'dGVzdCBjb25maWd1cmF0aW9u' | /usr/bin/base64 -d > /etc/dhcp/automation/test.conf",
                    print_progress_bars=False,
                ),
                mock.call("/usr/local/sbin/dhcpincludes -r commit", print_progress_bars=False),
            ]
        )

    def test_push_configuration_echo_fail(self):
        """Test push_configuration, where writing to the file fails."""
        config = get_mock_config()
        hosts = get_mock_suc_fail_hosts()
        hosts.run_sync.side_effect = list(hosts.run_sync.side_effect)[1:]
        self._setup_dhcp_mocks(hosts=hosts)

        with pytest.raises(dhcp.DHCPError, match="Failed to create snippet"):
            self.dhcp.push_configuration(config)

    def test_push_configuration_refresh_fail(self):
        """When the call to _refresh_dhcp() fails, it should remove the snippet and refresh again."""
        config = get_mock_config()
        hosts = get_mock_suc_fail_hosts()
        hosts.run_sync.side_effect = [
            "some value",
            RemoteExecutionError("mock error", 1, iter(())),
            "some value",
            "some value",
        ]
        self._setup_dhcp_mocks(hosts=hosts)

        with pytest.raises(dhcp.DHCPRestartError, match="Failed to refresh the DHCP server when running dhcpincludes"):
            self.dhcp.push_configuration(config)

        hosts.run_sync.assert_has_calls(
            [mock.call("/usr/local/sbin/dhcpincludes -r commit", print_progress_bars=False)]
        )

    def test_remove_config_ok(self):
        """Test remove_configuration where everything succeeds."""
        config = get_mock_config()
        hosts = self._setup_dhcp_mocks()
        with mock.patch("spicerack.dhcp.RemoteHosts") as mock_remotehosts:
            configsha256 = sha256(str(config).encode()).hexdigest()
            mock_remotehosts.results_to_list.return_value = [[None, f"{configsha256} {config.filename}"]]
            self.dhcp.remove_configuration(config)

        call_sha256 = mock.call(
            f"sha256sum {dhcp.DHCP_TARGET_PATH}/{config.filename}",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )
        call_rm = mock.call(
            f"/bin/rm -v {dhcp.DHCP_TARGET_PATH}/{config.filename}", print_output=False, print_progress_bars=False
        )
        call_refresh = mock.call("/usr/local/sbin/dhcpincludes -r commit", print_progress_bars=False)

        hosts.run_sync.assert_has_calls([call_sha256, call_rm, call_refresh])

    def test_remove_config_sha256_noresult(self):
        """Test remove_configuration where sha256sum returns nothing."""
        config = get_mock_config()
        self._setup_dhcp_mocks()
        with mock.patch("spicerack.dhcp.RemoteHosts") as mock_remotehosts:
            mock_remotehosts.results_to_list.return_value = []

            with pytest.raises(dhcp.DHCPError, match="No output when trying to checksum snippet"):
                self.dhcp.remove_configuration(config)

    def test_remove_config_sha256_fail(self):
        """Test remove_configuration where sha256sum fails to run."""
        config = get_mock_config()
        self._setup_dhcp_mocks(hosts=get_mock_fail_hosts())

        with pytest.raises(dhcp.DHCPError, match="Failed to checksum"):
            self.dhcp.remove_configuration(config)

    def test_remove_config_sha256_mismatch(self):
        """Test remove_configuration where sha256sum and the locally computed sum mismatch."""
        config = get_mock_config()
        self._setup_dhcp_mocks()
        configsha256 = sha256(str(config).encode()).hexdigest()
        config.__str__.return_value = "different test configuration"

        with mock.patch("spicerack.dhcp.RemoteHosts") as mock_remotehosts:
            mock_remotehosts.results_to_list.return_value = [[None, f"{configsha256} {config.filename}"]]
            with pytest.raises(dhcp.DHCPError, match="has a mismatched SHA256, refusing to remove it"):
                self.dhcp.remove_configuration(config)

    def test_remove_config_sha256_mismatch_force(self):
        """Test remove_configuration where there is a sha256 mismatch but we pass force=True."""
        config = get_mock_config()
        hosts = self._setup_dhcp_mocks()
        configsha256 = sha256(str(config).encode()).hexdigest()
        config.__str__.return_value = "different test configuration"

        with mock.patch("spicerack.dhcp.RemoteHosts") as mock_remotehosts:
            mock_remotehosts.results_to_list.return_value = [[None, f"{configsha256} {config.filename}"]]
            self.dhcp.remove_configuration(config, force=True)

        call_rm = mock.call(
            f"/bin/rm -v {dhcp.DHCP_TARGET_PATH}/{config.filename}", print_output=False, print_progress_bars=False
        )
        call_refresh = mock.call("/usr/local/sbin/dhcpincludes -r commit", print_progress_bars=False)

        hosts.run_sync.assert_has_calls([call_rm, call_refresh])

    def test_remove_config_rm_fail(self):
        """Test remove_configuration where rm fails."""
        config = get_mock_config()
        self._setup_dhcp_mocks(hosts=get_mock_suc_fail_hosts())

        with mock.patch("spicerack.dhcp.RemoteHosts") as mock_remotehosts:
            configsha256 = sha256(str(config).encode()).hexdigest()
            mock_remotehosts.results_to_list.return_value = [[None, f"{configsha256} {config.filename}"]]

            with pytest.raises(dhcp.DHCPError, match="Failed to remove snippet"):
                self.dhcp.remove_configuration(config)

    def test_push_context_manager(self):
        """Test push context manager success."""
        config = get_mock_config()
        hosts = self._setup_dhcp_mocks()
        call_sha256 = mock.call(
            f"sha256sum {dhcp.DHCP_TARGET_PATH}/{config.filename}",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )
        call_rm = mock.call(
            f"/bin/rm -v {dhcp.DHCP_TARGET_PATH}/{config.filename}", print_output=False, print_progress_bars=False
        )
        call_refresh = mock.call("/usr/local/sbin/dhcpincludes -r commit", print_progress_bars=False)

        with mock.patch("spicerack.dhcp.RemoteHosts") as mock_remotehosts:
            configsha256 = sha256(str(config).encode()).hexdigest()
            mock_remotehosts.results_to_list.return_value = [[None, f"{configsha256} {config.filename}"]]

            with self.dhcp.config(config):
                pass

        hosts.run_sync.assert_has_calls([call_sha256, call_rm, call_refresh])

    def test_push_context_manager_raise(self):
        """Test push context manager where internal code raises."""
        config = get_mock_config()
        hosts = self._setup_dhcp_mocks()

        call_write = mock.call(
            f"/bin/echo '{config.config_base64}' | /usr/bin/base64 -d > {dhcp.DHCP_TARGET_PATH}/{config.filename}",
            print_progress_bars=False,
        )
        call_sha256 = mock.call(
            f"sha256sum {dhcp.DHCP_TARGET_PATH}/{config.filename}",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )
        call_rm = mock.call(
            f"/bin/rm -v {dhcp.DHCP_TARGET_PATH}/{config.filename}", print_output=False, print_progress_bars=False
        )
        call_refresh = mock.call("/usr/local/sbin/dhcpincludes -r commit", print_progress_bars=False)

        with mock.patch("spicerack.dhcp.RemoteHosts") as mock_remotehosts:
            configsha256 = sha256(str(config).encode()).hexdigest()
            mock_remotehosts.results_to_list.return_value = [[None, f"{configsha256} {config.filename}"]]
            with pytest.raises(RuntimeError):
                with self.dhcp.config(config):
                    raise RuntimeError()

        hosts.run_sync.assert_has_calls([call_write, call_refresh, call_sha256, call_rm, call_refresh])



"""PeeringDB module."""
import json
import logging
import time
from collections.abc import MutableMapping
from pathlib import Path
from typing import Optional, cast

from wmflib.requests import http_session

from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


class PeeringDBError(SpicerackError):
    """Custom exception class for errors of the PeeringDB class."""


class CacheMiss(SpicerackError):
    """Custom exception class for cache management."""


class PeeringDB:
    """Basic dumb wrapper over the PeeringDB API.

    Implements the beta/v0 PeeringDB API. Tries to be smart by:

        a) keeping a persistent keep-alived session for multiple requests
        b) operating a local filesystem cache, if so desired.

    """

    baseurl: str = "https://www.peeringdb.com/api/"
    """The PeeringDB base API URL."""

    def __init__(
        self,
        *,
        ttl: int = 86400,
        cachedir: Optional[Path] = None,
        proxies: Optional[MutableMapping[str, str]] = None,
        token: str = "",
    ):
        """Initiliaze the module.

        Arguments:
            ttl: TTL for cached objects.
            cachedir: Root path for objects caching.
            proxies: Proxies for Internet access.
            token: PeeringDB read-only token.

        """
        self.session = http_session(".".join((self.__module__, self.__class__.__name__)))
        if token:
            self.session.headers.update({"Authorization": f"Api-Key {token}"})

        if proxies:
            self.session.proxies = proxies

        self.ttl = ttl
        self.use_cache = cachedir is not None and self.ttl > 0

        self.cachedir: Path
        if cachedir is not None:
            self.cachedir = cast(Path, cachedir)
            self.cachedir.mkdir(exist_ok=True)

    @staticmethod
    def _get_cache_key(resource: str, *, resource_id: Optional[int] = None, filters: Optional[dict] = None) -> str:
        """Return a cache key based on the resource requested.

        Arguments:
            resource: The PeeringDB resource requested.
            resource_id: Optional resource number.
            filters: A dictionary of addtional filter parameters.

        """
        if resource_id is None and filters is None:
            return f"{resource}/index"
        if filters is None:
            return f"{resource}/{str(resource_id)}"
        filter_key = "/".join("/".join([k, str(v)]) for k, v in filters.items())
        if resource_id is None:
            return f"{resource}/{filter_key}"
        return f"{resource}/{str(resource_id)}/{filter_key}"

    def fetch_asn(self, asn: int) -> dict:
        """Fetch a specific asn data.

        Arguments:
            asn: The Autonomous system number.

        """
        return self.fetch("net", filters={"asn": asn, "depth": 2})

    def fetch(self, resource: str, resource_id: Optional[int] = None, filters: Optional[dict] = None) -> dict:
        """Get a PeeringDB resource.

        Arguments:
            resource: The PeeringDB resource requested.
            resource_id: Optional resource number.
            filters: A dictionary of addtional filter parameters.

        """
        if resource_id is None:
            endpoint = resource
        else:
            endpoint = f"{resource}/{str(resource_id)}"

        cache_key = self._get_cache_key(resource=resource, resource_id=resource_id, filters=filters)
        try:
            json_response = self._cache_get(cache_key)
        except CacheMiss as e:
            url = self.baseurl + endpoint
            logging.debug("Fetching %s from PeeringDB (params: %s)", url, filters)
            raw_response = self.session.get(url, params=filters)
            if not raw_response.ok:
                raise PeeringDBError(
                    f"Server response with status {raw_response.status_code}" f" ({raw_response.text})"
                ) from e
            json_response = raw_response.json()
            self._cache_put(json_response, cache_key)
        return json_response["data"]

    def _cache_get(self, cache_key: str) -> dict:
        """Get the resource from the on disk cache if present.

        Arguments:
            cache_key: The on-disk path of the resource.

        """
        if not self.use_cache:
            raise CacheMiss()

        cachefile = self.cachedir / cache_key
        try:
            mtime = cachefile.stat().st_mtime
            age = time.time() - mtime
            if age > self.ttl:
                raise CacheMiss()
            return json.loads(cachefile.read_text())

        except OSError as e:
            raise CacheMiss() from e

    def _cache_put(self, content: dict, cache_key: str) -> None:
        """Write the resource to the on disk cache if configured to do so.

        Arguments:
            content: Dictionary of data to cache.
            cache_key: The on-disk path of the resource.

        """
        if not self.use_cache:
            return

        cachefile = self.cachedir / cache_key
        cachefile.parent.mkdir(exist_ok=True, parents=True)
        cachefile.write_text(json.dumps(content, indent=2))



"""Remote module to execute commands on hosts via Cumin."""
import logging
import math
import time
from collections.abc import Callable, Iterator, Sequence
from datetime import datetime, timedelta
from typing import Any, Optional, Union

from ClusterShell.MsgTree import MsgTreeElem
from cumin import Config, CuminError, NodeSet, query, transport, transports
from cumin.cli import target_batch_size
from cumin.transports import Command
from cumin.transports.clustershell import NullReporter, TqdmReporter

from spicerack.confctl import ConftoolEntity
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError

logger = logging.getLogger(__name__)


class RemoteError(SpicerackError):
    """Custom exception class for errors of this module."""


class RemoteCheckError(SpicerackCheckError):
    """Custom exception class for check errors of this module."""


class RemoteExecutionError(RemoteError):
    """Custom exception class for remote execution errors."""

    def __init__(self, retcode: int, message: str, results: Iterator[tuple[NodeSet, MsgTreeElem]]) -> None:
        """Override parent constructor to add the return code attribute.

        Arguments:
            retcode: the return code of the remote execution.
            message: the exception message.
            results: the cumin results.

        """
        super().__init__(f"{message} (exit_code={retcode})")
        self.retcode = retcode
        self.results = results


class RemoteClusterExecutionError(RemoteError):
    """Custom exception class for collecting multiple execution errors on a cluster."""

    def __init__(
        self,
        results: list[tuple[NodeSet, MsgTreeElem]],
        failures: list[RemoteExecutionError],
    ):
        """Override the parent constructor to add failures and results as attributes.

        Arguments:
            results: the remote results.
            failures: the list of exceptions raised in the cluster execution.

        """
        super().__init__(f"{len(failures)} hosts have failed execution")
        self.failures = failures
        self.results = results


class RemoteHostsAdapter:
    """Base adapter to write classes that expand the capabilities of RemoteHosts.

    This adapter class is a helper class to reduce duplication when writing classes that needs to add capabilities to a
    RemoteHosts instance. The goal is to not extend the RemoteHosts but instead delegate to its instances.
    This class fits when a single RemoteHosts instance is enough, but for more complex cases, in which multiple
    RemoteHosts instances should be orchestrated, it's ok to not extend this class and create a standalone one.
    """

    def __init__(self, remote_hosts: "RemoteHosts") -> None:
        """Initialize the instance.

        Arguments:
            remote_hosts: the instance to act on the remote hosts.

        """
        self._remote_hosts = remote_hosts

    def __str__(self) -> str:
        """String representation of the instance."""
        return str(self._remote_hosts)

    def __len__(self) -> int:
        """Get the number of target hosts in this instance."""
        return len(self._remote_hosts)


class LBRemoteCluster(RemoteHostsAdapter):
    """Class usable to operate on a cluster of servers with pooling/depooling logic in conftool."""

    def __init__(self, config: Config, remote_hosts: "RemoteHosts", conftool: ConftoolEntity) -> None:
        """Initialize the instance.

        Arguments:
            config: cumin configuration.
            remote_hosts: the instance to act on the remote hosts.
            conftool: the conftool entity to operate on.

        """
        self._config = config
        self._conftool = conftool
        super().__init__(remote_hosts)

    def run(
        self,
        *commands: Union[str, Command],
        svc_to_depool: Optional[list[str]] = None,
        batch_size: int = 1,
        batch_sleep: Optional[float] = None,
        is_safe: bool = False,
        max_failed_batches: int = 0,
        print_output: bool = True,
        print_progress_bars: bool = True,
    ) -> list[tuple[NodeSet, MsgTreeElem]]:
        """Run commands while depooling servers in groups of batch_size.

        For clusters behind a load balancer, we typically want to be able to
        depool a server from a specific service, then run any number of commands
        on it, and finally repool it.

        We also want to ensure we can only have at max N hosts depooled at any
        time. Given cumin doesn't have pre- and post- execution hooks, we break
        the remote run in smaller groups and execute on one group at a time,
        in parallel on all the servers.
        Note this works a bit differently than how the cumin moving window works,
        as here we'll have to wait for the execution on all servers in a group
        before moving on to the next.

        Arguments:
            *commands: Arbitrary number of commands to execute.
            svc_to_depool: A list of services (in conftool) to depool.
            batch_size: the batch size for cumin, as an integer. Defaults to 1.
            batch_sleep: the batch sleep in seconds to use before scheduling the next batch of hosts.
            is_safe: whether the command is safe to run also in dry-run mode because it's a read-only
            max_failed_batches: Maximum number of batches that can fail. Defaults to 0.
            print_output: whether to print Cumin's output to stdout.
            print_progress_bars: whether to print Cumin's progress bars to stderr.

        Returns:
            What :py:meth:`cumin.transports.BaseWorker.get_results` returns to allow to iterate over the results.

        Raises:
            spicerack.remote.RemoteExecutionError, spicerack.remote.RemoteClusterExecutionError: if the Cumin execution
                returns a non-zero exit code.

        """
        n_hosts = len(self._remote_hosts)
        # Ensure the batch size is smaller than the whole cluster. When acting on a cluster
        # we should never act on all hosts in parallel.
        # If that's needed, RemoteHosts can be used directly.
        # TODO: the right thing to do would be to check that batch_size is smaller than
        # (1 - depool_threshold) * pooled_hosts, but that would also need to know the current
        # state of the cluster.
        if batch_size <= 0 or batch_size >= n_hosts:
            raise RemoteError(f"Values for batch_size must be 0 < x < {n_hosts}, got {batch_size}")

        # If no service needs depooling, the standard behavior of remote_hosts.run_async is used
        # TODO: add the ability to select all services.
        if svc_to_depool is None:
            return list(
                self._remote_hosts.run_async(
                    *commands,
                    success_threshold=(1.0 - (max_failed_batches * batch_size) / n_hosts),
                    batch_size=batch_size,
                    batch_sleep=batch_sleep,
                    is_safe=is_safe,
                    print_output=print_output,
                    print_progress_bars=print_progress_bars,
                )
            )

        results = []
        # Find how much we must split the pool up to achieve the desired batch size.
        n_slices = math.ceil(n_hosts / batch_size)
        # TODO: add better failure handling. Right now we just support counting batches with a failure, while we might
        # want to support success_threshold.
        failures = []
        for remotes_slice in self._remote_hosts.split(n_slices):
            # Select the pooled servers for the selected services, from the group we're operating on now.
            with self._conftool.change_and_revert(
                "pooled",
                "yes",
                "no",
                service="|".join(svc_to_depool),
                name="|".join(remotes_slice.hosts.striter()),
            ):
                try:
                    for result in remotes_slice.run_async(
                        *commands,
                        is_safe=is_safe,
                        success_threshold=1.0,
                        print_output=print_output,
                        print_progress_bars=print_progress_bars,
                    ):
                        results.append(result)
                except RemoteExecutionError as e:
                    failures.append(e)
                    # Break the execution loop if more than the maximum number of failures happened.
                    if len(failures) > max_failed_batches:
                        break
            # TODO: skip sleep on the last run
            if batch_sleep is not None:
                time.sleep(batch_sleep)
        if failures:
            raise RemoteClusterExecutionError(results, failures)
        return results

    def restart_services(
        self,
        services: list[str],
        svc_to_depool: list[str],
        *,
        batch_size: int = 1,
        batch_sleep: Optional[float] = None,
        verbose: bool = True,
    ) -> list[tuple[NodeSet, MsgTreeElem]]:
        """Restart services in batches, removing the host from all the affected services first.

        Arguments:
            services: A list of services to act upon.
            svc_to_depool: A list of services (in conftool) to depool.
            batch_size: the batch size for cumin, as an integer. Defaults to 1.
            batch_sleep: the batch sleep between groups of runs.
            verbose: whether to print Cumin's output and progress bars to stdout/stderr.

        Returns:
            What :py:meth:`cumin.transports.BaseWorker.get_results` returns to allow to iterate over the results.

        Raises:
            spicerack.remote.RemoteExecutionError, spicerack.remote.RemoteClusterExecutionError: if the Cumin execution
                returns a non-zero exit code.

        """
        return self._act_on_services(
            services, svc_to_depool, "restart", batch_size, batch_sleep=batch_sleep, verbose=verbose
        )

    def reload_services(
        self,
        services: list[str],
        svc_to_depool: list[str],
        *,
        batch_size: int = 1,
        batch_sleep: Optional[float] = None,
        verbose: bool = True,
    ) -> list[tuple[NodeSet, MsgTreeElem]]:
        """Reload services in batches, removing the host from all the affected services first.

        Arguments:
            services: A list of services to act upon.
            svc_to_depool: A list of services (in conftool) to depool.
            batch_size: the batch size for cumin, as an integer.Defaults to 1
            batch_sleep: the batch sleep between groups of runs.
            verbose: whether to print Cumin's output and progress bars to stdout/stderr.

        Returns:
            What :py:meth:`cumin.transports.BaseWorker.get_results` returns to allow to iterate over the results.

        Raises:
            spicerack.remote.RemoteExecutionError, spicerack.remote.RemoteClusterExecutionError: if the Cumin execution
                returns a non-zero exit code.

        """
        return self._act_on_services(
            services, svc_to_depool, "reload", batch_size, batch_sleep=batch_sleep, verbose=verbose
        )

    def _act_on_services(  # pylint: disable=too-many-arguments
        self,
        services: list[str],
        svc_to_depool: list[str],
        what: str,
        batch_size: int,
        batch_sleep: Optional[float] = None,
        verbose: bool = True,
    ) -> list[tuple[NodeSet, MsgTreeElem]]:
        """Act on services in batches, depooling the servers first.

        Arguments:
            services: A list of services to act upon.
            svc_to_depool: A list of services (in conftool) to depool.
            what: Action to perform. restart by default.
            batch_size: the batch size for cumin, as an integer.
            batch_sleep: the batch sleep between groups of runs.
            verbose: whether to print Cumin's output and progress bars to stdout/stderr.

        Returns:
            What :py:meth:`cumin.transports.BaseWorker.get_results` returns to allow to iterate over the results.

        Raises:
            spicerack.remote.RemoteExecutionError, spicerack.remote.RemoteClusterExecutionError: if the Cumin execution
                returns a non-zero exit code.

        """
        commands = [f'systemctl {what} "{svc}"' for svc in services]
        return self.run(
            *commands,
            svc_to_depool=svc_to_depool,
            batch_size=batch_size,
            batch_sleep=batch_sleep,
            is_safe=False,
            print_output=verbose,
            print_progress_bars=verbose,
        )


class Remote:
    """Remote class to interact with Cumin."""

    def __init__(self, config: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            config: the path of Cumin's configuration file.
            dry_run: whether this is a DRY-RUN.

        """
        self._config = Config(config)
        self._dry_run = dry_run

    def query(self, query_string: str, use_sudo: bool = False) -> "RemoteHosts":
        """Execute a Cumin query and return the matching hosts.

        Arguments:
            query_string: the Cumin query string to execute.
            use_sudo: If True will prepend 'sudo -i' to every command.

        """
        # TODO: Revisit the current implementation of sudo once Cumin has native support for it.
        try:
            hosts = query.Query(self._config).execute(query_string)
        except CuminError as e:
            raise RemoteError("Failed to execute Cumin query") from e

        return RemoteHosts(self._config, hosts, dry_run=self._dry_run, use_sudo=use_sudo)

    def query_confctl(self, conftool: ConftoolEntity, **tags: str) -> LBRemoteCluster:
        """Execute a conftool node query and return the matching hosts.

        Arguments:
            conftool: the conftool instance for the node type objects.
            tags: Conftool tags for node type objects as keyword arguments.

        Raises:
            spicerack.remote.RemoteError: if unable to query the hosts.

        """
        # get the list of hosts from confctl
        try:
            hosts_conftool = [obj.name for obj in conftool.get(**tags)]
            query_string = ",".join(hosts_conftool)
        except SpicerackError as e:
            raise RemoteError("Failed to execute the conftool query") from e

        remote_hosts = self.query(query_string)
        host_diff = set(hosts_conftool) - set(remote_hosts.hosts)
        if host_diff:  # pragma: no cover | This should never happen with a direct backend.
            logger.warning("Hosts present in conftool but not in puppet: %s", ",".join(host_diff))
        return LBRemoteCluster(self._config, remote_hosts, conftool)


class RemoteHosts:
    """Remote Executor class.

    This class can be extended to customize the interaction with remote hosts passing a custom factory function to
    `spicerack.remote.Remote.query`.
    """

    def __init__(self, config: Config, hosts: NodeSet, dry_run: bool = True, use_sudo: bool = False) -> None:
        """Initialize the instance.

        Arguments:
            config: the configuration for Cumin.
            hosts: the hosts to target for the remote execution.
            dry_run: whether this is a DRY-RUN.
            use_sudo: if True will prepend ``sudo -i`` to every command.

        Raises:
            spicerack.remote.RemoteError: if no hosts were provided.

        """
        if not hosts:
            raise RemoteError("No hosts provided")

        self._config = config
        self._hosts = hosts
        self._dry_run = dry_run
        self._use_sudo = use_sudo

    @property
    def hosts(self) -> NodeSet:
        """Returns a copy of the current hosts targeted."""
        return self._hosts.copy()

    def __str__(self) -> str:
        """String representation of the instance."""
        return str(self._hosts)

    def __len__(self) -> int:
        """Returns the number of hosts targeted."""
        return len(self._hosts)

    def split(self, n_slices: int) -> Iterator["RemoteHosts"]:
        """Split the current remote in n_slices RemoteHosts instances.

        Arguments:
            n_slices: the number of slices to slice the remote in.

        Yields:
            spicerack.remote.RemoteHosts: the instances for the subset of nodes.

        """
        for nodeset in self._hosts.split(n_slices):
            yield RemoteHosts(self._config, nodeset, dry_run=self._dry_run, use_sudo=self._use_sudo)

    def get_subset(self, subset: NodeSet) -> "RemoteHosts":
        """Return a new RemoteHosts instance with a subset of the existing set of hosts.

        Arguments:
            subset: the subset of hosts to select. They must all be part of the current set.

        Raises:
            spicerack.remote.RemoteError: if any host in the subset is not part of the instance hosts.

        Returns:
            a new instance with only the subset hosts.

        """
        if not self._hosts.issuperset(subset):
            raise RemoteError(f"The provided set {subset} is not a subset of the current set {self._hosts}")

        return RemoteHosts(self._config, subset, dry_run=self._dry_run, use_sudo=self._use_sudo)

    @staticmethod
    def _prepend_sudo(command: Union[str, Command]) -> Union[str, Command]:
        if isinstance(command, str):
            return "sudo -i " + command

        return Command(
            "sudo -i " + command.command,
            timeout=command.timeout,
            ok_codes=command.ok_codes,
        )

    def run_async(
        self,
        *commands: Union[str, Command],
        success_threshold: float = 1.0,
        batch_size: Optional[Union[int, str]] = None,
        batch_sleep: Optional[float] = None,
        is_safe: bool = False,
        print_output: bool = True,
        print_progress_bars: bool = True,
    ) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Execute commands on hosts matching a query via Cumin in async mode and return its results.

        Arguments:
            *commands: arbitrary number of commands to execute on the target hosts.
            success_threshold: to consider the execution successful, must be between 0.0 and 1.0.
            batch_size: the batch size for cumin, either as percentage (e.g. ``25%``) or absolute number (e.g. ``5``).
            batch_sleep: the batch sleep in seconds to use in Cumin before scheduling the next host.
            is_safe: whether the command is safe to run also in dry-run mode because it's a read-only command that
                doesn't modify the state.
            print_output: whether to print Cumin's output to stdout.
            print_progress_bars: whether to print Cumin's progress bars to stderr.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        return self._execute(
            list(commands),
            mode="async",
            success_threshold=success_threshold,
            batch_size=batch_size,
            batch_sleep=batch_sleep,
            is_safe=is_safe,
            print_output=print_output,
            print_progress_bars=print_progress_bars,
        )

    def run_sync(
        self,
        *commands: Union[str, Command],
        success_threshold: float = 1.0,
        batch_size: Optional[Union[int, str]] = None,
        batch_sleep: Optional[float] = None,
        is_safe: bool = False,
        print_output: bool = True,
        print_progress_bars: bool = True,
    ) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Execute commands on hosts matching a query via Cumin in sync mode and returns its results.

        Arguments:
            *commands: arbitrary number of commands to execute on the target hosts.
            success_threshold: to consider the execution successful, must be between 0.0 and 1.0.
            batch_size: the batch size for cumin, either as percentage (e.g. ``25%``) or absolute number (e.g. ``5``).
            batch_sleep: the batch sleep in seconds to use in Cumin before scheduling the next host.
            is_safe: whether the command is safe to run also in dry-run mode because it's a read-only command that
                doesn't modify the state.
            print_output: whether to print Cumin's output to stdout.
            print_progress_bars: whether to print Cumin's progress bars to stderr.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        return self._execute(
            list(commands),
            mode="sync",
            success_threshold=success_threshold,
            batch_size=batch_size,
            batch_sleep=batch_sleep,
            is_safe=is_safe,
            print_output=print_output,
            print_progress_bars=print_progress_bars,
        )

    def reboot(self, batch_size: int = 1, batch_sleep: Optional[float] = 180.0) -> None:
        """Reboot hosts.

        Arguments:
            batch_size: how many hosts to reboot in parallel.
            batch_sleep: how long to sleep between one reboot and the next.

        """
        if len(self._hosts) == 1:  # Temporary workaround until T213296 is fixed.
            batch_sleep = None

        logger.info(
            "Rebooting %d hosts in batches of %d with %.1fs of sleep in between: %s",
            len(self._hosts),
            batch_size,
            batch_sleep if batch_sleep is not None else 0.0,
            self._hosts,
        )

        self.run_sync(
            transports.Command("reboot-host", timeout=30),
            batch_size=batch_size,
            batch_sleep=batch_sleep,
        )

    @retry(
        tries=240,
        delay=timedelta(seconds=10),
        backoff_mode="constant",
        exceptions=(RemoteCheckError,),
    )
    def wait_reboot_since(self, since: datetime, print_progress_bars: bool = True) -> None:
        """Poll the host until is reachable and has an uptime lower than the provided datetime.

        Arguments:
            since: the time after which the host should have booted.
            print_progress_bars: whether to print Cumin's progress bars to stderr.

        Raises:
            spicerack.remote.RemoteCheckError: if unable to connect to the host or the uptime is higher than expected.
                When in DRY-RUN mode, it will raise only if unable to connect.

        """
        delta = (datetime.utcnow() - since).total_seconds()
        try:
            uptimes = self.uptime(print_progress_bars=print_progress_bars)
        except (RemoteExecutionError, RemoteError) as e:
            if self._dry_run:
                raise RemoteCheckError(f"Reboot for {self._hosts} not found: unable to get uptime") from e
            raise RemoteCheckError(
                f"Reboot for {self._hosts} not found yet, keep polling for it: unable to get uptime"
            ) from e

        for nodeset, uptime in uptimes:
            if uptime >= delta:
                msg = f"uptime {round(uptime, 2)} > threshold {round(delta, 2)}"
                if self._dry_run:
                    logger.info(
                        "Reboot for %s not found - expected in dry run mode, host not rebooted: %s", nodeset, msg
                    )
                else:
                    raise RemoteCheckError(f"Reboot for {nodeset} not found yet, keep polling for it: {msg}")

        if not self._dry_run:
            logger.info("Found reboot since %s for hosts %s", since, self._hosts)

    def uptime(self, print_progress_bars: bool = True) -> list[tuple[NodeSet, float]]:
        """Get current uptime.

        Arguments:
            print_progress_bars: whether to print Cumin's progress bars to stderr.

        Returns:
            A list of 2-element :py:class:`tuple` instances with hosts :py:class:`ClusterShell.NodeSet.NodeSet`
            as first item and :py:class:`float` uptime as second item.

        Raises:
            spicerack.remote.RemoteError: if unable to parse the output as an uptime.

        """
        results = self.run_sync(
            transports.Command("cat /proc/uptime", timeout=10),
            is_safe=True,
            print_output=False,
            print_progress_bars=print_progress_bars,
        )
        logger.debug("Got uptime for hosts %s", self._hosts)
        # Callback to extract the uptime from /proc/uptime (i.e. getting 12345.67 from '12345.67 123456789.00').
        return RemoteHosts.results_to_list(results, callback=lambda output: float(output.split()[0]))

    @staticmethod
    def results_to_list(
        results: Iterator[tuple[NodeSet, MsgTreeElem]],
        callback: Optional[Callable] = None,
    ) -> list[tuple[NodeSet, Any]]:
        """Extract execution results into a list converting them with an optional callback.

        Todo:
            move it directly into Cumin.

        Arguments:
            results: generator returned by run_sync() and run_async() to iterate over the results.
            callback: an optional callable to apply to each result output (it can be multiline).
                The callback will be called with a the string output as the only parameter and must return the
                extracted value. The return type can be chosen freely.

        Returns:
            A list of 2-element tuples with hosts :py:class:`ClusterShell.NodeSet.NodeSet` as first item and the
            extracted outputs :py:class:`str` as second. This is because NodeSet are not hashable.

        Raises:
            spicerack.remote.RemoteError: if unable to run the callback.

        """
        extracted = []
        for nodeset, output in results:
            result = output.message().decode().strip()
            if callback is not None:
                try:
                    result = callback(result)
                except Exception as e:
                    raise RemoteError(
                        f"Unable to extract data with {callback.__name__} for {nodeset} from: {result}"
                    ) from e

            extracted.append((nodeset, result))

        return extracted

    def _execute(  # pylint: disable=too-many-arguments
        self,
        commands: Sequence[Union[str, Command]],
        mode: str = "sync",
        success_threshold: float = 1.0,
        batch_size: Optional[Union[int, str]] = None,
        batch_sleep: Optional[float] = None,
        is_safe: bool = False,
        print_output: bool = True,
        print_progress_bars: bool = True,
    ) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Lower level Cumin's execution of commands on the target nodes and return its results.

        Arguments:
            commands: the list of commands to execute on the target hosts, either a list of commands or a list
                of cumin.transports.Command instances.
            mode: the Cumin's mode of execution. Accepted values: sync, async.
            success_threshold: to consider the execution successful, must be between 0.0 and 1.0.
            batch_size: the batch size for cumin, either as percentage (e.g. ``25%``) or absolute number (e.g. ``5``).
            batch_sleep: the batch sleep in seconds to use in Cumin before scheduling the next host.
            is_safe: whether the command is safe to run also in dry-run mode because it's a read-only command that
                doesn't modify the state.
            print_output: whether to print Cumin's output to stdout.
            print_progress_bars: whether to print Cumin's progress bars to stderr.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        if batch_size is None:
            parsed_batch_size = {"value": None, "ratio": None}
        else:
            parsed_batch_size = target_batch_size(str(batch_size))

        if self._use_sudo:
            commands = [self._prepend_sudo(command) for command in commands]

        target = transports.Target(
            self._hosts,
            batch_size=parsed_batch_size["value"],
            batch_size_ratio=parsed_batch_size["ratio"],
            batch_sleep=batch_sleep,
        )
        worker = transport.Transport.new(self._config, target)
        worker.commands = commands
        worker.handler = mode
        worker.success_threshold = success_threshold
        worker.progress_bars = print_progress_bars
        if print_output:
            worker.reporter = TqdmReporter
        else:
            worker.reporter = NullReporter

        logger.debug(
            "Executing commands %s on %d hosts: %s",
            commands,
            len(target.hosts),
            str(target.hosts),
        )

        if self._dry_run and not is_safe:
            return iter(())  # Empty generator

        ret = worker.execute()

        if ret != 0 and not self._dry_run:
            raise RemoteExecutionError(ret, "Cumin execution failed", worker.get_results())

        return worker.get_results()
